# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name,redefined-builtin
# pylint: disable=super-init-not-called

"""Module providing definitions of Pulse Qobj classes."""

import copy
import pprint
from typing import Union, List

import numpy
from qiskit.qobj.common import QobjDictField
from qiskit.qobj.common import QobjHeader
from qiskit.qobj.common import QobjExperimentHeader
from qiskit.utils import deprecate_func


class QobjMeasurementOption:
    """An individual measurement option."""

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(self, name, params=None):
        """Instantiate a new QobjMeasurementOption object.

        Args:
            name (str): The name of the measurement option
            params (list): The parameters of the measurement option.
        """
        self.name = name
        if params is not None:
            self.params = params

    def to_dict(self):
        """Return a dict format representation of the QobjMeasurementOption.

        Returns:
            dict: The dictionary form of the QasmMeasurementOption.
        """
        out_dict = {"name": self.name}
        if hasattr(self, "params"):
            out_dict["params"] = self.params
        return out_dict

    @classmethod
    def from_dict(cls, data):
        """Create a new QobjMeasurementOption object from a dictionary.

        Args:
            data (dict): A dictionary for the experiment config

        Returns:
            QobjMeasurementOption: The object from the input dictionary.
        """
        name = data.pop("name")
        return cls(name, **data)

    def __eq__(self, other):
        if isinstance(other, QobjMeasurementOption):
            if self.to_dict() == other.to_dict():
                return True
        return False


class PulseQobjInstruction:
    """A class representing a single instruction in an PulseQobj Experiment."""

    _COMMON_ATTRS = [
        "ch",
        "conditional",
        "val",
        "phase",
        "frequency",
        "duration",
        "qubits",
        "memory_slot",
        "register_slot",
        "label",
        "type",
        "pulse_shape",
        "parameters",
    ]

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(
        self,
        name,
        t0,
        ch=None,
        conditional=None,
        val=None,
        phase=None,
        duration=None,
        qubits=None,
        memory_slot=None,
        register_slot=None,
        kernels=None,
        discriminators=None,
        label=None,
        type=None,
        pulse_shape=None,
        parameters=None,
        frequency=None,
    ):
        """Instantiate a new PulseQobjInstruction object.

        Args:
            name (str): The name of the instruction
            t0 (int): Pulse start time in integer **dt** units.
            ch (str): The channel to apply the pulse instruction.
            conditional (int): The register to use for a conditional for this
                instruction
            val (complex): Complex value to apply, bounded by an absolute value
                of 1.
            phase (float): if a ``fc`` instruction, the frame change phase in
                radians.
            frequency (float): if a ``sf`` instruction, the frequency in Hz.
            duration (int): The duration of the pulse in **dt** units.
            qubits (list): A list of ``int`` representing the qubits the
                instruction operates on
            memory_slot (list): If a ``measure`` instruction this is a list
                of ``int`` containing the list of memory slots to store the
                measurement results in (must be the same length as qubits).
                If a ``bfunc`` instruction this is a single ``int`` of the
                memory slot to store the boolean function result in.
            register_slot (list): If a ``measure`` instruction this is a list
                of ``int`` containing the list of register slots in which to
                store the measurement results (must be the same length as
                qubits). If a ``bfunc`` instruction this is a single ``int``
                of the register slot in which to store the result.
            kernels (list): List of :class:`QobjMeasurementOption` objects
                defining the measurement kernels and set of parameters if the
                measurement level is 1 or 2. Only used for ``acquire``
                instructions.
            discriminators (list): A list of :class:`QobjMeasurementOption`
                used to set the discriminators to be used if the measurement
                level is 2. Only used for ``acquire`` instructions.
            label (str): Label of instruction
            type (str): Type of instruction
            pulse_shape (str): The shape of the parametric pulse
            parameters (dict): The parameters for a parametric pulse
        """
        self.name = name
        self.t0 = t0
        if ch is not None:
            self.ch = ch
        if conditional is not None:
            self.conditional = conditional
        if val is not None:
            self.val = val
        if phase is not None:
            self.phase = phase
        if frequency is not None:
            self.frequency = frequency
        if duration is not None:
            self.duration = duration
        if qubits is not None:
            self.qubits = qubits
        if memory_slot is not None:
            self.memory_slot = memory_slot
        if register_slot is not None:
            self.register_slot = register_slot
        if kernels is not None:
            self.kernels = kernels
        if discriminators is not None:
            self.discriminators = discriminators
        if label is not None:
            self.label = label
        if type is not None:
            self.type = type
        if pulse_shape is not None:
            self.pulse_shape = pulse_shape
        if parameters is not None:
            self.parameters = parameters

    def to_dict(self):
        """Return a dictionary format representation of the Instruction.

        Returns:
            dict: The dictionary form of the PulseQobjInstruction.
        """
        out_dict = {"name": self.name, "t0": self.t0}
        for attr in self._COMMON_ATTRS:
            if hasattr(self, attr):
                out_dict[attr] = getattr(self, attr)
        if hasattr(self, "kernels"):
            out_dict["kernels"] = [x.to_dict() for x in self.kernels]
        if hasattr(self, "discriminators"):
            out_dict["discriminators"] = [x.to_dict() for x in self.discriminators]
        return out_dict

    def __repr__(self):
        out = f'PulseQobjInstruction(name="{self.name}", t0={self.t0}'
        for attr in self._COMMON_ATTRS:
            attr_val = getattr(self, attr, None)
            if attr_val is not None:
                if isinstance(attr_val, str):
                    out += f', {attr}="{attr_val}"'
                else:
                    out += f", {attr}={attr_val}"
        out += ")"
        return out

    def __str__(self):
        out = f"Instruction: {self.name}\n"
        out += f"\t\tt0: {self.t0}\n"
        for attr in self._COMMON_ATTRS:
            if hasattr(self, attr):
                out += f"\t\t{attr}: {getattr(self, attr)}\n"
        return out

    @classmethod
    def from_dict(cls, data):
        """Create a new PulseQobjExperimentConfig object from a dictionary.

        Args:
            data (dict): A dictionary for the experiment config

        Returns:
            PulseQobjInstruction: The object from the input dictionary.
        """
        schema = {
            "discriminators": QobjMeasurementOption,
            "kernels": QobjMeasurementOption,
        }
        skip = ["t0", "name"]

        # Pulse instruction data is nested dictionary.
        # To avoid deepcopy and avoid mutating the source object, create new dict here.
        in_data = {}
        for key, value in data.items():
            if key in skip:
                continue
            if key == "parameters":
                # This is flat dictionary of parametric pulse parameters
                formatted_value = value.copy()
                if "amp" in formatted_value:
                    formatted_value["amp"] = _to_complex(formatted_value["amp"])
                in_data[key] = formatted_value
                continue
            if key in schema:
                if isinstance(value, list):
                    in_data[key] = list(map(schema[key].from_dict, value))
                else:
                    in_data[key] = schema[key].from_dict(value)
            else:
                in_data[key] = value

        return cls(data["name"], data["t0"], **in_data)

    def __eq__(self, other):
        if isinstance(other, PulseQobjInstruction):
            if self.to_dict() == other.to_dict():
                return True
        return False


def _to_complex(value: Union[List[float], complex]) -> complex:
    """Convert the input value to type ``complex``.
    Args:
        value: Value to be converted.
    Returns:
        Input value in ``complex``.
    Raises:
        TypeError: If the input value is not in the expected format.
    """
    if isinstance(value, list) and len(value) == 2:
        return complex(value[0], value[1])
    elif isinstance(value, complex):
        return value

    raise TypeError(f"{value} is not in a valid complex number format.")


class PulseQobjConfig(QobjDictField):
    """A configuration for a Pulse Qobj."""

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(
        self,
        meas_level,
        meas_return,
        pulse_library,
        qubit_lo_freq,
        meas_lo_freq,
        memory_slot_size=None,
        rep_time=None,
        rep_delay=None,
        shots=None,
        seed_simulator=None,
        memory_slots=None,
        **kwargs,
    ):
        """Instantiate a PulseQobjConfig object.

        Args:
            meas_level (int): The measurement level to use.
            meas_return (int): The level of measurement information to return.
            pulse_library (list): A list of :class:`PulseLibraryItem` objects
                which define the set of primitive pulses
            qubit_lo_freq (list): List of frequencies (as floats) for the qubit
                driver LO's in GHz.
            meas_lo_freq (list): List of frequencies (as floats) for the'
                measurement driver LO's in GHz.
            memory_slot_size (int): Size of each memory slot if the output is
                Level 0.
            rep_time (int): Time per program execution in sec. Must be from the list provided
                by the backend (``backend.configuration().rep_times``). Defaults to the first entry
                in ``backend.configuration().rep_times``.
            rep_delay (float): Delay between programs in sec. Only supported on certain
                backends (``backend.configuration().dynamic_reprate_enabled`` ). If supported,
                ``rep_delay`` will be used instead of ``rep_time`` and must be from the range
                supplied by the backend (``backend.configuration().rep_delay_range``). Default is
                ``backend.configuration().default_rep_delay``.
            shots (int): The number of shots
            seed_simulator (int): the seed to use in the simulator
            memory_slots (list): The number of memory slots on the device
            kwargs: Additional free form key value fields to add to the
                configuration
        """
        self.meas_level = meas_level
        self.meas_return = meas_return
        self.pulse_library = pulse_library
        self.qubit_lo_freq = qubit_lo_freq
        self.meas_lo_freq = meas_lo_freq
        if memory_slot_size is not None:
            self.memory_slot_size = memory_slot_size
        if rep_time is not None:
            self.rep_time = rep_time
        if rep_delay is not None:
            self.rep_delay = rep_delay
        if shots is not None:
            self.shots = int(shots)

        if seed_simulator is not None:
            self.seed_simulator = int(seed_simulator)

        if memory_slots is not None:
            self.memory_slots = int(memory_slots)

        if kwargs:
            self.__dict__.update(kwargs)

    def to_dict(self):
        """Return a dictionary format representation of the Pulse Qobj config.

        Returns:
            dict: The dictionary form of the PulseQobjConfig.
        """
        out_dict = copy.copy(self.__dict__)
        if hasattr(self, "pulse_library"):
            out_dict["pulse_library"] = [x.to_dict() for x in self.pulse_library]

        return out_dict

    @classmethod
    def from_dict(cls, data):
        """Create a new PulseQobjConfig object from a dictionary.

        Args:
            data (dict): A dictionary for the config

        Returns:
            PulseQobjConfig: The object from the input dictionary.
        """
        if "pulse_library" in data:
            pulse_lib = data.pop("pulse_library")
            pulse_lib_obj = [PulseLibraryItem.from_dict(x) for x in pulse_lib]
            data["pulse_library"] = pulse_lib_obj
        return cls(**data)


class PulseQobjExperiment:
    """A Pulse Qobj Experiment.

    Each instance of this class is used to represent an individual Pulse
    experiment as part of a larger Pulse Qobj.
    """

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(self, instructions, config=None, header=None):
        """Instantiate a PulseQobjExperiment.

        Args:
            config (PulseQobjExperimentConfig): A config object for the experiment
            header (PulseQobjExperimentHeader): A header object for the experiment
            instructions (list): A list of :class:`PulseQobjInstruction` objects
        """
        if config is not None:
            self.config = config
        if header is not None:
            self.header = header
        self.instructions = instructions

    def to_dict(self):
        """Return a dictionary format representation of the Experiment.

        Returns:
            dict: The dictionary form of the PulseQobjExperiment.
        """
        out_dict = {"instructions": [x.to_dict() for x in self.instructions]}
        if hasattr(self, "config"):
            out_dict["config"] = self.config.to_dict()
        if hasattr(self, "header"):
            out_dict["header"] = self.header.to_dict()
        return out_dict

    def __repr__(self):
        instructions_str = [repr(x) for x in self.instructions]
        instructions_repr = "[" + ", ".join(instructions_str) + "]"
        out = "PulseQobjExperiment("
        out += instructions_repr
        if hasattr(self, "config") or hasattr(self, "header"):
            out += ", "
        if hasattr(self, "config"):
            out += "config=" + str(repr(self.config)) + ", "
        if hasattr(self, "header"):
            out += "header=" + str(repr(self.header)) + ", "
        out += ")"
        return out

    def __str__(self):
        out = "\nPulse Experiment:\n"
        if hasattr(self, "config"):
            config = pprint.pformat(self.config.to_dict())
        else:
            config = "{}"
        if hasattr(self, "header"):
            header = pprint.pformat(self.header.to_dict() or {})
        else:
            header = "{}"
        out += f"Header:\n{header}\n"
        out += f"Config:\n{config}\n\n"
        for instruction in self.instructions:
            out += f"\t{instruction}\n"
        return out

    @classmethod
    def from_dict(cls, data):
        """Create a new PulseQobjExperiment object from a dictionary.

        Args:
            data (dict): A dictionary for the experiment config

        Returns:
            PulseQobjExperiment: The object from the input dictionary.
        """
        config = None
        if "config" in data:
            config = PulseQobjExperimentConfig.from_dict(data.pop("config"))
        header = None
        if "header" in data:
            header = QobjExperimentHeader.from_dict(data.pop("header"))
        instructions = None
        if "instructions" in data:
            instructions = [
                PulseQobjInstruction.from_dict(inst) for inst in data.pop("instructions")
            ]
        return cls(instructions, config, header)

    def __eq__(self, other):
        if isinstance(other, PulseQobjExperiment):
            if self.to_dict() == other.to_dict():
                return True
        return False


class PulseQobjExperimentConfig(QobjDictField):
    """A config for a single Pulse experiment in the qobj."""

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(self, qubit_lo_freq=None, meas_lo_freq=None, **kwargs):
        """Instantiate a PulseQobjExperimentConfig object.

        Args:
            qubit_lo_freq (List[float]): List of qubit LO frequencies in GHz.
            meas_lo_freq (List[float]): List of meas readout LO frequencies in GHz.
            kwargs: Additional free form key value fields to add to the configuration
        """
        if qubit_lo_freq is not None:
            self.qubit_lo_freq = qubit_lo_freq
        if meas_lo_freq is not None:
            self.meas_lo_freq = meas_lo_freq
        if kwargs:
            self.__dict__.update(kwargs)


class PulseLibraryItem:
    """An item in a pulse library."""

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(self, name, samples):
        """Instantiate a pulse library item.

        Args:
            name (str): A name for the pulse.
            samples (list[complex]): A list of complex values defining pulse
                shape.
        """
        self.name = name
        if isinstance(samples[0], list):
            self.samples = numpy.array([complex(sample[0], sample[1]) for sample in samples])
        else:
            self.samples = samples

    def to_dict(self):
        """Return a dictionary format representation of the pulse library item.

        Returns:
            dict: The dictionary form of the PulseLibraryItem.
        """
        return {"name": self.name, "samples": self.samples}

    @classmethod
    def from_dict(cls, data):
        """Create a new PulseLibraryItem object from a dictionary.

        Args:
            data (dict): A dictionary for the experiment config

        Returns:
            PulseLibraryItem: The object from the input dictionary.
        """
        return cls(**data)

    def __repr__(self):
        return f"PulseLibraryItem({self.name}, {repr(self.samples)})"

    def __str__(self):
        return f"Pulse Library Item:\n\tname: {self.name}\n\tsamples: {self.samples}"

    def __eq__(self, other):
        if isinstance(other, PulseLibraryItem):
            if self.to_dict() == other.to_dict():
                return True
        return False


class PulseQobj:
    """A Pulse Qobj."""

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The `Qobj` class and related functionality are part of the deprecated "
        "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires `Qobj` it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
    )
    def __init__(self, qobj_id, config, experiments, header=None):
        """Instantiate a new Pulse Qobj Object.

        Each Pulse Qobj object is used to represent a single payload that will
        be passed to a Qiskit provider. It mirrors the Qobj the published
        `Qobj specification <https://arxiv.org/abs/1809.03452>`_ for Pulse
        experiments.

        Args:
            qobj_id (str): An identifier for the qobj
            config (PulseQobjConfig): A config for the entire run
            header (QobjHeader): A header for the entire run
            experiments (list): A list of lists of :class:`PulseQobjExperiment`
                objects representing an experiment
        """
        self.qobj_id = qobj_id
        self.config = config
        self.header = header or QobjHeader()
        self.experiments = experiments
        self.type = "PULSE"
        self.schema_version = "1.2.0"

    def __repr__(self):
        experiments_str = [repr(x) for x in self.experiments]
        experiments_repr = "[" + ", ".join(experiments_str) + "]"
        return (
            f"PulseQobj(qobj_id='{self.qobj_id}', config={repr(self.config)}, "
            f"experiments={experiments_repr}, header={repr(self.header)})"
        )

    def __str__(self):
        out = f"Pulse Qobj: {self.qobj_id}:\n"
        config = pprint.pformat(self.config.to_dict())
        out += f"Config: {str(config)}\n"
        header = pprint.pformat(self.header.to_dict())
        out += f"Header: {str(header)}\n"
        out += "Experiments:\n"
        for experiment in self.experiments:
            out += str(experiment)
        return out

    def to_dict(self):
        """Return a dictionary format representation of the Pulse Qobj.

        Note this dict is not in the json wire format expected by IBMQ and qobj
        specification because complex numbers are still of type complex. Also
        this may contain native numpy arrays. When serializing this output
        for use with IBMQ you can leverage a json encoder that converts these
        as expected. For example:

        .. code-block::

            import json
            import numpy

            class QobjEncoder(json.JSONEncoder):
                def default(self, obj):
                    if isinstance(obj, numpy.ndarray):
                        return obj.tolist()
                    if isinstance(obj, complex):
                        return (obj.real, obj.imag)
                    return json.JSONEncoder.default(self, obj)

            json.dumps(qobj.to_dict(), cls=QobjEncoder)

        Returns:
            dict: A dictionary representation of the PulseQobj object
        """
        out_dict = {
            "qobj_id": self.qobj_id,
            "header": self.header.to_dict(),
            "config": self.config.to_dict(),
            "schema_version": self.schema_version,
            "type": self.type,
            "experiments": [x.to_dict() for x in self.experiments],
        }
        return out_dict

    @classmethod
    def from_dict(cls, data):
        """Create a new PulseQobj object from a dictionary.

        Args:
            data (dict): A dictionary representing the PulseQobj to create. It
                will be in the same format as output by :func:`to_dict`.

        Returns:
            PulseQobj: The PulseQobj from the input dictionary.
        """
        config = None
        if "config" in data:
            config = PulseQobjConfig.from_dict(data["config"])
        experiments = None
        if "experiments" in data:
            experiments = [PulseQobjExperiment.from_dict(exp) for exp in data["experiments"]]
        header = None
        if "header" in data:
            header = QobjHeader.from_dict(data["header"])

        return cls(
            qobj_id=data.get("qobj_id"), config=config, experiments=experiments, header=header
        )

    def __eq__(self, other):
        if isinstance(other, PulseQobj):
            if self.to_dict() == other.to_dict():
                return True
        return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Reference instruction that is a placeholder for subroutine."""
from __future__ import annotations

from qiskit.circuit.parameterexpression import ParameterExpression
from qiskit.pulse.channels import Channel
from qiskit.pulse.exceptions import PulseError, UnassignedReferenceError
from qiskit.pulse.instructions import instruction
from qiskit.utils.deprecate_pulse import deprecate_pulse_func


class Reference(instruction.Instruction):
    """Pulse compiler directive that refers to a subroutine.

    If a pulse program uses the same subset of instructions multiple times, then
    using the :class:`~.Reference` class may significantly reduce the memory footprint of
    the program. This instruction only stores the set of strings to identify the subroutine.

    The actual pulse program can be stored in the :attr:`ScheduleBlock.references` of the
    :class:`.ScheduleBlock` that this reference instruction belongs to.

    You can later assign schedules with the :meth:`ScheduleBlock.assign_references` method.
    This allows you to build the main program without knowing the actual subroutine,
    that is supplied at a later time.
    """

    # Delimiter for representing nested scope.
    scope_delimiter = "::"

    # Delimiter for tuple keys.
    key_delimiter = ","

    @deprecate_pulse_func
    def __init__(self, name: str, *extra_keys: str):
        """Create new reference.

        Args:
            name: Name of subroutine.
            extra_keys: Optional. A set of string keys that may be necessary to
                refer to a particular subroutine. For example, when we use
                "sx" as a name to refer to the subroutine of an sx pulse,
                this name might be used among schedules for different qubits.
                In this example, you may specify "q0" in the extra keys
                to distinguish the sx schedule for qubit 0 from others.
                The user can use an arbitrary number of extra string keys to
                uniquely determine the subroutine.
        """
        # Run validation
        ref_keys = (name,) + tuple(extra_keys)
        super().__init__(operands=ref_keys, name=name)

    def _validate(self):
        """Called after initialization to validate instruction data.

        Raises:
            PulseError: When a key is not a string.
            PulseError: When a key in ``ref_keys`` contains the scope delimiter.
        """
        for key in self.ref_keys:
            if not isinstance(key, str):
                raise PulseError(f"Keys must be strings. '{repr(key)}' is not a valid object.")
            if self.scope_delimiter in key or self.key_delimiter in key:
                raise PulseError(
                    f"'{self.scope_delimiter}' and '{self.key_delimiter}' are reserved. "
                    f"'{key}' is not a valid key string."
                )

    @property
    def ref_keys(self) -> tuple[str, ...]:
        """Returns unique key of the subroutine."""
        return self.operands

    @property
    def duration(self) -> int | ParameterExpression:
        """Duration of this instruction."""
        raise UnassignedReferenceError(f"Subroutine is not assigned to {self.ref_keys}.")

    @property
    def channels(self) -> tuple[Channel, ...]:
        """Returns the channels that this schedule uses."""
        raise UnassignedReferenceError(f"Subroutine is not assigned to {self.ref_keys}.")

    @property
    def parameters(self) -> set:
        """Parameters which determine the instruction behavior."""
        return set()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.key_delimiter.join(self.ref_keys)})"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Stabilizer to circuit function
"""
from __future__ import annotations

from collections.abc import Collection

import numpy as np

from qiskit.quantum_info import PauliList
from qiskit.exceptions import QiskitError
from qiskit.circuit import QuantumCircuit
from qiskit.quantum_info.operators.symplectic.clifford import Clifford


def synth_circuit_from_stabilizers(
    stabilizers: Collection[str],
    allow_redundant: bool = False,
    allow_underconstrained: bool = False,
    invert: bool = False,
) -> QuantumCircuit:
    # pylint: disable=line-too-long
    """Synthesis of a circuit that generates a state stabilized by the stabilizers
    using Gaussian elimination with Clifford gates.
    If the stabilizers are underconstrained, and ``allow_underconstrained`` is ``True``,
    the circuit will output one of the states stabilized by the stabilizers.
    Based on stim implementation.

    Args:
        stabilizers: List of stabilizer strings
        allow_redundant: Allow redundant stabilizers (i.e., some stabilizers
            can be products of the others)
        allow_underconstrained: Allow underconstrained set of stabilizers (i.e.,
            the stabilizers do not specify a unique state)
        invert: Return inverse circuit

    Returns:
        A circuit that generates a state stabilized by ``stabilizers``.

    Raises:
        QiskitError: if the stabilizers are invalid, do not commute, or contradict each other,
                     if the list is underconstrained and ``allow_underconstrained`` is ``False``,
                     or if the list is redundant and ``allow_redundant`` is ``False``.

    References:
        1. https://github.com/quantumlib/Stim/blob/c0dd0b1c8125b2096cd54b6f72884a459e47fe3e/src/stim/stabilizers/conversions.inl#L469
        2. https://quantumcomputing.stackexchange.com/questions/12721/how-to-calculate-destabilizer-group-of-toric-and-other-codes

    """
    stabilizer_list = PauliList(stabilizers)
    if np.any(stabilizer_list.phase % 2):
        raise QiskitError("Some stabilizers have an invalid phase")
    if len(stabilizer_list.commutes_with_all(stabilizer_list)) < len(stabilizer_list):
        raise QiskitError("Some stabilizers do not commute.")

    num_qubits = stabilizer_list.num_qubits
    circuit = QuantumCircuit(num_qubits)

    used = 0
    for i, stabilizer in enumerate(stabilizer_list):
        curr_stab = stabilizer.evolve(Clifford(circuit), frame="s")

        # Find pivot.
        pivot = used
        while pivot < num_qubits:
            if curr_stab[pivot].x or curr_stab[pivot].z:
                break
            pivot += 1

        if pivot == num_qubits:
            if curr_stab.x.any():
                raise QiskitError(
                    f"Stabilizer {i} ({stabilizer}) anti-commutes with some of "
                    "the previous stabilizers."
                )
            if curr_stab.phase == 2:
                raise QiskitError(
                    f"Stabilizer {i} ({stabilizer}) contradicts "
                    "some of the previous stabilizers."
                )
            if curr_stab.z.any() and not allow_redundant:
                raise QiskitError(
                    f"Stabilizer {i} ({stabilizer}) is a product of the others "
                    "and allow_redundant is False. Add allow_redundant=True "
                    "to the function call if you want to allow redundant stabilizers."
                )
            continue

        # Change pivot basis to the Z axis.
        if curr_stab[pivot].x:
            if curr_stab[pivot].z:
                circuit.h(pivot)
                circuit.s(pivot)
                circuit.h(pivot)
                circuit.s(pivot)
                circuit.s(pivot)
            else:
                circuit.h(pivot)

        # Cancel other terms in Pauli string.
        for j in range(num_qubits):
            if j == pivot or not (curr_stab[j].x or curr_stab[j].z):
                continue
            p = curr_stab[j].x + curr_stab[j].z * 2
            if p == 1:  # X
                circuit.h(pivot)
                circuit.cx(pivot, j)
                circuit.h(pivot)
            elif p == 2:  # Z
                circuit.cx(j, pivot)
            elif p == 3:  # Y
                circuit.h(pivot)
                circuit.s(j)
                circuit.s(j)
                circuit.s(j)
                circuit.cx(pivot, j)
                circuit.h(pivot)
                circuit.s(j)

        # Move pivot to diagonal.
        if pivot != used:
            circuit.swap(pivot, used)

        # fix sign
        curr_stab = stabilizer.evolve(Clifford(circuit), frame="s")
        if curr_stab.phase == 2:
            circuit.x(used)
        used += 1

    if used < num_qubits and not allow_underconstrained:
        raise QiskitError(
            "Stabilizers are underconstrained and allow_underconstrained is False."
            " Add allow_underconstrained=True  to the function call "
            "if you want to allow underconstrained stabilizers."
        )
    if invert:
        return circuit
    return circuit.inverse()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Double-CNOT gate."""

from qiskit.circuit.singleton import SingletonGate, stdlib_singleton_key
from qiskit.circuit.quantumregister import QuantumRegister
from qiskit.circuit._utils import with_gate_array
from qiskit._accelerate.circuit import StandardGate


@with_gate_array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 1, 0, 0], [0, 0, 1, 0]])
class DCXGate(SingletonGate):
    r"""Double-CNOT gate.

    A 2-qubit Clifford gate consisting of two back-to-back
    CNOTs with alternate controls.

    Can be applied to a :class:`~qiskit.circuit.QuantumCircuit`
    with the :meth:`~qiskit.circuit.QuantumCircuit.dcx` method.

    .. code-block:: text

                  ┌───┐
        q_0: ──■──┤ X ├
             ┌─┴─┐└─┬─┘
        q_1: ┤ X ├──■──
             └───┘

    This is a classical logic gate, equivalent to a CNOT-SWAP (CNS) sequence,
    and locally equivalent to an iSWAP.

    .. math::

        DCX\ q_0, q_1 =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0
            \end{pmatrix}
    """

    _standard_gate = StandardGate.DCXGate

    def __init__(self, label=None, *, duration=None, unit="dt"):
        """Create new DCX gate."""
        super().__init__("dcx", 2, [], label=label, duration=duration, unit=unit)

    _singleton_lookup_key = stdlib_singleton_key()

    def _define(self):
        """
        gate dcx a, b { cx a, b; cx b, a; }
        """
        # pylint: disable=cyclic-import
        from qiskit.circuit.quantumcircuit import QuantumCircuit
        from .x import CXGate

        q = QuantumRegister(2, "q")
        qc = QuantumCircuit(q, name=self.name)
        rules = [(CXGate(), [q[0], q[1]], []), (CXGate(), [q[1], q[0]], [])]
        for instr, qargs, cargs in rules:
            qc._append(instr, qargs, cargs)

        self.definition = qc

    def __eq__(self, other):
        return isinstance(other, DCXGate)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Z2Symmetries for SparsePauliOp."""

from __future__ import annotations

import itertools
from collections.abc import Iterable
from copy import deepcopy
import math
from typing import Union, cast

import numpy as np

from qiskit.exceptions import QiskitError
from ..operators import Pauli, SparsePauliOp


class Z2Symmetries:
    r"""
    The $Z_2$ symmetry converter identifies symmetries from the problem hamiltonian and uses them to
    provide a tapered - more efficient - representation of operators as Paulis for this problem. For each
    identified symmetry, one qubit can be eliminated in the Pauli representation at the cost of having to
    test two symmetry sectors (for the two possible eigenvalues - tapering values - of the symmetry).
    In certain problems such as the finding of the main operator's ground state, one can a priori
    identify the symmetry sector of the solution and thus effectively reduce the computational overhead.

    The following attributes can be read and updated once the ``Z2Symmetries`` object has been
    constructed.

    Attributes:
        tapering_values (list[int] or None): Values determining the sector.
        tol (float): The tolerance threshold for ignoring real and complex parts of a coefficient.

    References:
        [1]: Bravyi, S., et al, "Tapering off qubits to simulate fermionic Hamiltonians"
            `arXiv:1701.08213 <https://arxiv.org/abs/1701.08213>`__
    """

    def __init__(
        self,
        symmetries: Iterable[Pauli],
        sq_paulis: Iterable[Pauli],
        sq_list: Iterable[int],
        tapering_values: Iterable[int] | None = None,
        *,
        tol: float = 1e-14,
    ):
        r"""
        Args:
            symmetries: Object representing the list of $Z_2$ symmetries. These correspond to
                the generators of the symmetry group $\langle \tau_1, \tau_2\dots \rangle>$.
            sq_paulis: Object representing the list of single-qubit Pauli $\sigma^x_{q(i)}$
                anti-commuting with the symmetry $\tau_i$ and commuting with all the other symmetries
                $\tau_{j\neq i}$. These operators are used to construct the unitary Clifford operators.
            sq_list: The list of indices $q(i)$ of the single-qubit Pauli operators used to build the
                Clifford operators.
            tapering_values: List of eigenvalues determining the symmetry sector for each symmetry.
            tol: Tolerance threshold for ignoring real and complex parts of a coefficient.

        Raises:
            QiskitError: Invalid paulis. The lists of symmetries, single-qubit paulis support paulis
                and tapering values must be of equal length. This length is the number of applied
                symmetries and translates directly to the number of eliminated qubits.
        """
        symmetries = list(symmetries)
        sq_paulis = list(sq_paulis)
        sq_list = list(sq_list)
        tapering_values = None if tapering_values is None else list(tapering_values)

        if len(symmetries) != len(sq_paulis):
            raise QiskitError(
                f"The number of Z2 symmetries, {len(symmetries)}, has to match the number \
                of single-qubit pauli operators, {len(sq_paulis)}."
            )

        if len(sq_paulis) != len(sq_list):
            raise QiskitError(
                f"The number of single-qubit pauli operators, {len(sq_paulis)}, has to match the length \
                of the of single-qubit list, {len(sq_list)}."
            )

        if tapering_values is not None:
            if len(sq_list) != len(tapering_values):
                raise QiskitError(
                    f"The length of the single-qubit list, {len(sq_list)}, must match the length of the \
                    tapering values, {len(tapering_values)} ."
                )

        self._symmetries = symmetries
        self._sq_paulis = sq_paulis
        self._sq_list = sq_list
        self.tapering_values = tapering_values
        self.tol = tol

    @property
    def symmetries(self) -> list[Pauli]:
        """Return symmetries."""
        return self._symmetries

    @property
    def sq_paulis(self) -> list[Pauli]:
        """Return sq paulis."""
        return self._sq_paulis

    @property
    def cliffords(self) -> list[SparsePauliOp]:
        """
        Get clifford operators, built based on symmetries and single-qubit X.

        Returns:
            A list of unitaries used to diagonalize the Hamiltonian.
        """
        cliffords = [
            (SparsePauliOp(pauli_symm) + SparsePauliOp(sq_pauli)) / math.sqrt(2)
            for pauli_symm, sq_pauli in zip(self._symmetries, self._sq_paulis)
        ]
        return cliffords

    @property
    def sq_list(self) -> list[int]:
        """Return sq list."""
        return self._sq_list

    @property
    def settings(self) -> dict:
        """Return operator settings."""
        return {
            "symmetries": self._symmetries,
            "sq_paulis": self._sq_paulis,
            "sq_list": self._sq_list,
            "tapering_values": self.tapering_values,
        }

    def __str__(self):
        ret = ["Z2 symmetries:"]
        ret.append("Symmetries:")
        for symmetry in self._symmetries:
            ret.append(symmetry.to_label())
        ret.append("Single-Qubit Pauli X:")
        for x in self._sq_paulis:
            ret.append(x.to_label())
        ret.append("Cliffords:")
        for c in self.cliffords:
            ret.append(str(c))
        ret.append("Qubit index:")
        ret.append(str(self._sq_list))
        ret.append("Tapering values:")
        if self.tapering_values is None:
            possible_values = [
                str(list(coeff)) for coeff in itertools.product([1, -1], repeat=len(self._sq_list))
            ]
            possible_values = ", ".join(x for x in possible_values)
            ret.append("  - Possible values: " + possible_values)
        else:
            ret.append(str(self.tapering_values))

        ret = "\n".join(ret)
        return ret

    def is_empty(self) -> bool:
        """
        Check the z2_symmetries is empty or not.

        Returns:
            Empty or not.
        """
        return len(self._symmetries) == 0 or len(self._sq_paulis) == 0 or len(self._sq_list) == 0

    @classmethod
    def find_z2_symmetries(cls, operator: SparsePauliOp) -> Z2Symmetries:
        """
        Finds Z2 Pauli-type symmetries of a :class:`.SparsePauliOp`.

        Returns:
            A ``Z2Symmetries`` instance.
        """
        pauli_symmetries = []
        sq_paulis = []
        sq_list = []

        stacked_paulis = []

        test_idx = {
            "X_or_I": [(0, 0), (1, 0)],
            "Y_or_I": [(0, 0), (1, 1)],
            "Z_or_I": [(0, 0), (0, 1)],
        }
        test_row = {
            "Z_or_I": [(1, 0), (1, 1)],
            "X_or_I": [(0, 1), (1, 1)],
            "Y_or_I": [(0, 1), (1, 0)],
        }

        pauli_bool = {
            "Z_or_I": [False, True],
            "X_or_I": [True, False],
            "Y_or_I": [True, True],
        }

        if _sparse_pauli_op_is_zero(operator):
            return cls([], [], [], None)

        for pauli in iter(operator):
            stacked_paulis.append(
                np.concatenate((pauli.paulis.x[0], pauli.paulis.z[0]), axis=0).astype(int)
            )

        stacked_matrix = np.stack(stacked_paulis)
        symmetries = _kernel_f2(stacked_matrix)

        if not symmetries:
            return cls([], [], [], None)

        stacked_symmetries = np.stack(symmetries)
        symm_shape = stacked_symmetries.shape
        half_symm_shape = symm_shape[1] // 2
        stacked_symm_del = [
            np.delete(stacked_symmetries, row, axis=0) for row in range(symm_shape[0])
        ]

        def _test_symmetry_row_col(row: int, col: int, idx_test: list, row_test: list) -> bool:
            """
            Utility method that determines how to build the list of single-qubit Pauli X operators and
            the list of corresponding qubit indices from the stacked symmetries.
            This method is successively applied to Z type, X type and Y type symmetries (in this order)
            to build the letter at position (col) of the Pauli word corresponding to the symmetry at
            position (row).

            Args:
                row (int): Index of the symmetry for which the single-qubit Pauli X operator is being
                    built.
                col (int): Index of the letter in the Pauli word corresponding to the single-qubit Pauli
                    X operator.
                idx_test (list): List of possibilities for the stacked symmetries at all other rows
                    than row.
                row_test (list): List of possibilities for the stacked symmetries at row.

            Returns:
                Whether or not this symmetry type should be used to build this letter of this
                single-qubit Pauli X operator.
            """
            stacked_symm_idx_tests = np.array(
                [
                    (
                        stacked_symm_del[row][symm_idx, col],
                        stacked_symm_del[row][symm_idx, col + half_symm_shape],
                    )
                    in idx_test
                    for symm_idx in range(symm_shape[0] - 1)
                ]
            )

            stacked_symm_row_test = (
                stacked_symmetries[row, col],
                stacked_symmetries[row, col + half_symm_shape],
            ) in row_test

            return bool(np.all(stacked_symm_idx_tests)) and stacked_symm_row_test

        for row in range(symm_shape[0]):
            pauli_symmetries.append(
                Pauli(
                    (
                        stacked_symmetries[row, :half_symm_shape],
                        stacked_symmetries[row, half_symm_shape:],
                    )
                )
            )
            # Try all cases for the symmetries other than row: Z or I, Y or I, X or I on col qubit.
            # One test will return true.
            # Build the single-qubit Pauli accordingly.
            # Build the index list accordingly.
            for col in range(half_symm_shape):
                for key in ("Z_or_I", "X_or_I", "Y_or_I"):
                    current_test_result = _test_symmetry_row_col(
                        row, col, test_idx[key], test_row[key]
                    )
                    if current_test_result:
                        sq_paulis.append(
                            Pauli((np.zeros(half_symm_shape), np.zeros(half_symm_shape)))
                        )
                        sq_paulis[row].z[col] = pauli_bool[key][0]
                        sq_paulis[row].x[col] = pauli_bool[key][1]
                        sq_list.append(col)
                        break
                if current_test_result:
                    # We break out of the loop over columns only when one valid test is identified.
                    break

        return cls(pauli_symmetries, sq_paulis, sq_list, None)

    def convert_clifford(self, operator: SparsePauliOp) -> SparsePauliOp:
        """This method operates the first part of the tapering.
        It converts the operator by composing it with the clifford unitaries defined in the current
        symmetry.

        Args:
            operator: The to-be-tapered operator.

        Returns:
            ``SparsePauliOp`` corresponding to the converted operator.

        """

        if not self.is_empty() and not _sparse_pauli_op_is_zero(operator):
            # If the operator is zero then we can skip the following.
            for clifford in self.cliffords:
                operator = cast(SparsePauliOp, clifford @ operator @ clifford)
                operator = operator.simplify(atol=0.0)

        return operator

    def taper_clifford(self, operator: SparsePauliOp) -> Union[SparsePauliOp, list[SparsePauliOp]]:
        """Operate the second part of the tapering.
        This function assumes that the input operators have already been transformed using
        :meth:`convert_clifford`. The redundant qubits due to the symmetries are dropped and
        replaced by their two possible eigenvalues.

        Args:
            operator: Partially tapered operator resulting from a call to :meth:`convert_clifford`.

        Returns:
            If tapering_values is None: [:class:`SparsePauliOp`]; otherwise, :class:`SparsePauliOp`.

        """

        tapered_ops: Union[SparsePauliOp, list[SparsePauliOp]]
        if self.is_empty():
            tapered_ops = operator
        else:
            # If the operator is zero we still need to taper the operator to reduce its size i.e. the
            # number of qubits so for example 0*"IIII" could taper to 0*"II" when symmetries remove
            # two qubits.
            if self.tapering_values is None:
                tapered_ops = [
                    self._taper(operator, list(coeff))
                    for coeff in itertools.product([1, -1], repeat=len(self._sq_list))
                ]
            else:
                tapered_ops = self._taper(operator, self.tapering_values)

        return tapered_ops

    def taper(self, operator: SparsePauliOp) -> Union[SparsePauliOp, list[SparsePauliOp]]:
        """
        Taper an operator based on the z2_symmetries info and sector defined by `tapering_values`.
        Returns operator if the symmetry object is empty.

        The tapering is a two-step algorithm which first converts the operator into a
        :class:`SparsePauliOp` with same eigenvalues but where some qubits are only acted upon
        with the Pauli operators I or X.
        The number M of these redundant qubits is equal to the number M of identified symmetries.

        The second step of the reduction consists in replacing these qubits with the possible
        eigenvalues of the corresponding Pauli X, giving 2^M new operators with M less qubits.
        If an eigenvalue sector was previously identified for the solution, then this reduces to
        1 new operator with M less qubits.

        Args:
            operator: The to-be-tapered operator.

        Returns:
            If tapering_values is None: [:class:`SparsePauliOp`]; otherwise, :class:`SparsePauliOp`.

        """

        converted_ops = self.convert_clifford(operator)
        tapered_ops = self.taper_clifford(converted_ops)

        return tapered_ops

    def _taper(self, op: SparsePauliOp, curr_tapering_values: list[int]) -> SparsePauliOp:
        pauli_list = []
        for pauli_term in iter(op):
            coeff_out = pauli_term.coeffs[0]
            for idx, qubit_idx in enumerate(self._sq_list):
                if pauli_term.paulis.z[0, qubit_idx] or pauli_term.paulis.x[0, qubit_idx]:
                    coeff_out = curr_tapering_values[idx] * coeff_out
            z_temp = np.delete(pauli_term.paulis.z[0].copy(), np.asarray(self._sq_list))
            x_temp = np.delete(pauli_term.paulis.x[0].copy(), np.asarray(self._sq_list))
            pauli_list.append((Pauli((z_temp, x_temp)).to_label(), coeff_out))

        spo = SparsePauliOp.from_list(pauli_list).simplify(atol=0.0)
        spo = spo.chop(self.tol)
        return spo

    def __eq__(self, other: Z2Symmetries) -> bool:
        """
        Overload `==` operation to evaluate equality between Z2Symmetries.

        Args:
            other: The `Z2Symmetries` to compare to self.

        Returns:
            A bool equal to the equality of self and other.
        """
        if not isinstance(other, Z2Symmetries):
            return False

        return (
            self.symmetries == other.symmetries
            and self.sq_paulis == other.sq_paulis
            and self.sq_list == other.sq_list
            and self.tapering_values == other.tapering_values
        )


def _kernel_f2(matrix_in):
    """
    Compute the kernel of a binary matrix on the binary finite field.

    Args:
        matrix_in (numpy.ndarray): Binary matrix.

    Returns:
        The list of kernel vectors.
    """
    size = matrix_in.shape
    kernel = []
    matrix_in_id = np.vstack((matrix_in, np.identity(size[1])))
    matrix_in_id_ech = (_row_echelon_f2(matrix_in_id.transpose())).transpose()

    for col in range(size[1]):
        if np.array_equal(
            matrix_in_id_ech[0 : size[0], col], np.zeros(size[0])
        ) and not np.array_equal(matrix_in_id_ech[size[0] :, col], np.zeros(size[1])):
            kernel.append(matrix_in_id_ech[size[0] :, col])

    return kernel


def _row_echelon_f2(matrix_in):
    """
    Compute the row Echelon form of a binary matrix on the binary finite field.

    Args:
        matrix_in (numpy.ndarray): Binary matrix.

    Returns:
        Matrix_in in Echelon row form.
    """
    size = matrix_in.shape

    for i in range(size[0]):
        pivot_index = 0
        for j in range(size[1]):
            if matrix_in[i, j] == 1:
                pivot_index = j
                break
        for k in range(size[0]):
            if k != i and matrix_in[k, pivot_index] == 1:
                matrix_in[k, :] = np.mod(matrix_in[k, :] + matrix_in[i, :], 2)

    matrix_out_temp = deepcopy(matrix_in)
    indices = []
    matrix_out = np.zeros(size)

    for i in range(size[0] - 1):
        if np.array_equal(matrix_out_temp[i, :], np.zeros(size[1])):
            indices.append(i)
    for row in np.sort(indices)[::-1]:
        matrix_out_temp = np.delete(matrix_out_temp, (row), axis=0)

    matrix_out[0 : size[0] - len(indices), :] = matrix_out_temp
    matrix_out = matrix_out.astype(int)

    return matrix_out


def _sparse_pauli_op_is_zero(op: SparsePauliOp) -> bool:
    """Returns whether or not this operator represents a zero operation."""
    op = op.simplify()
    return len(op.coeffs) == 1 and op.coeffs[0] == 0



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Look-up table for variable parameters in QuantumCircuit.
"""

from collections.abc import MappingView


class ParameterView(MappingView):
    """Temporary class to transition from a set return-type to list.

    Derives from a list but implements all set methods, but all set-methods emit deprecation
    warnings.
    """

    def __init__(self, iterable=None):
        if iterable is not None:
            self.data = list(iterable)
        else:
            self.data = []

        super().__init__(self.data)

    def copy(self):
        """Copy the ParameterView."""
        return self.__class__(self.data.copy())

    def isdisjoint(self, x):
        """Check whether self and the input are disjoint."""
        return not any(element in self for element in x)

    def remove(self, x):
        """Remove an existing element from the view."""
        self.data.remove(x)

    def __repr__(self):
        """Format the class as string."""
        return f"ParameterView({self.data})"

    def __getitem__(self, index):
        """Get items."""
        return self.data[index]

    def __and__(self, x):
        """Get the intersection between self and the input."""
        inter = []
        for element in self:
            if element in x:
                inter.append(element)

        return self.__class__(inter)

    def __rand__(self, x):
        """Get the intersection between self and the input."""
        return self.__and__(x)

    def __iand__(self, x):
        """Get the intersection between self and the input in-place."""
        for element in self:
            if element not in x:
                self.remove(element)
        return self

    def __len__(self):
        """Get the length."""
        return len(self.data)

    def __or__(self, x):
        """Get the union of self and the input."""
        return set(self) | set(x)

    def __sub__(self, x):
        """Get the difference between self and the input."""
        return set(self) - set(x)

    def __xor__(self, x):
        """Get the symmetric difference between self and the input."""
        return set(self) ^ set(x)

    def __ne__(self, other):
        return set(other) != set(self)

    def __eq__(self, other):
        return set(other) == set(self)

    def __le__(self, x):
        return all(element in x for element in self)

    def __lt__(self, x):
        if x != self:
            return self <= x
        return False

    def __ge__(self, x):
        return all(element in self for element in x)

    def __gt__(self, x):
        if x != self:
            return self >= x
        return False

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, x):
        return x in self.data

    __hash__: None  # type: ignore
    __ror__ = __or__



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Tolerances mixin class.
"""

from abc import ABCMeta
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators.predicates import ATOL_DEFAULT, RTOL_DEFAULT


class TolerancesMeta(ABCMeta):
    """Metaclass to handle tolerances"""

    def __init__(cls, *args, **kwargs):
        cls._ATOL_DEFAULT = ATOL_DEFAULT
        cls._RTOL_DEFAULT = RTOL_DEFAULT
        cls._MAX_TOL = 1e-4
        super().__init__(cls, args, kwargs)

    def _check_value(cls, value, value_name):
        """Check if value is within valid ranges"""
        if value < 0:
            raise QiskitError(f"Invalid {value_name} ({value}) must be non-negative.")
        if value > cls._MAX_TOL:
            raise QiskitError(f"Invalid {value_name} ({value}) must be less than {cls._MAX_TOL}.")

    @property
    def atol(cls):
        """Default absolute tolerance parameter for float comparisons."""
        return cls._ATOL_DEFAULT

    @atol.setter
    def atol(cls, value):
        """Set default absolute tolerance parameter for float comparisons."""
        cls._check_value(value, "atol")  # pylint: disable=no-value-for-parameter
        cls._ATOL_DEFAULT = value

    @property
    def rtol(cls):
        """Default relative tolerance parameter for float comparisons."""
        return cls._RTOL_DEFAULT

    @rtol.setter
    def rtol(cls, value):
        """Set default relative tolerance parameter for float comparisons."""
        cls._check_value(value, "rtol")  # pylint: disable=no-value-for-parameter
        cls._RTOL_DEFAULT = value


class TolerancesMixin(metaclass=TolerancesMeta):
    """Mixin Class for tolerances"""

    @property
    def atol(self):
        """Default absolute tolerance parameter for float comparisons."""
        return self.__class__.atol

    @property
    def rtol(self):
        """Default relative tolerance parameter for float comparisons."""
        return self.__class__.rtol



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests preset pass managers with 1Q backend"""

from ddt import ddt

from qiskit import QuantumCircuit
from qiskit.compiler import transpile
from qiskit.providers.fake_provider import Fake1Q, GenericBackendV2
from qiskit.providers.basic_provider import BasicSimulator
from qiskit.transpiler import TranspilerError
from test import combine  # pylint: disable=wrong-import-order
from test import QiskitTestCase  # pylint: disable=wrong-import-order

Fake1QV2 = GenericBackendV2(
    num_qubits=1, basis_gates=["u1", "u2", "u3"], coupling_map=None, dtm=1.3333, seed=42
)


def emptycircuit():
    """Empty circuit"""
    return QuantumCircuit()


def circuit_3516():
    """Circuit from https://github.com/Qiskit/qiskit-terra/issues/3516 should fail"""
    circuit = QuantumCircuit(2, 1)
    circuit.h(0)
    circuit.ry(0.11, 1)
    circuit.measure([0], [0])
    return circuit


@ddt
class Test1QFailing(QiskitTestCase):
    """1Q tests that should fail."""

    @combine(
        circuit=[circuit_3516],
        level=[0, 1, 2, 3],
        dsc="Transpiling {circuit.__name__} at level {level} should fail",
        name="{circuit.__name__}_level{level}_fail_v1",
    )
    def test(self, circuit, level):
        """All the levels with all the 1Q backendV1"""
        with self.assertRaises(TranspilerError):
            with self.assertWarns(DeprecationWarning):
                transpile(circuit(), backend=Fake1Q(), optimization_level=level, seed_transpiler=42)


@ddt
class Test1QV2Failing(QiskitTestCase):
    """1QV2 tests that should fail."""

    @combine(
        circuit=[circuit_3516],
        level=[0, 1, 2, 3],
        dsc="Transpiling {circuit.__name__} at level {level} should fail",
        name="{circuit.__name__}_level{level}_fail_v2",
    )
    def test(self, circuit, level):
        """All the levels with all the 1Q backendV2"""
        with self.assertRaises(TranspilerError):
            transpile(circuit(), backend=Fake1QV2, optimization_level=level, seed_transpiler=42)


@ddt
class Test1QWorking(QiskitTestCase):
    """1QV1 tests that should work."""

    @combine(
        circuit=[emptycircuit],
        level=[0, 1, 2, 3],
        dsc="Transpiling {circuit.__name__} at level {level} should work",
        name="{circuit.__name__}_level{level}_valid_v1",
    )
    def test_device(self, circuit, level):
        """All the levels with all the 1Q backendV1"""
        with self.assertWarns(DeprecationWarning):
            result = transpile(
                circuit(), backend=Fake1Q(), optimization_level=level, seed_transpiler=42
            )
        self.assertIsInstance(result, QuantumCircuit)


@ddt
class TestBasicSimulatorWorking(QiskitTestCase):
    """All the levels with a simulator backend"""

    @combine(
        circuit=[circuit_3516],
        level=[0, 1, 2, 3],
        dsc="Transpiling {circuit.__name__} at level {level} should work for simulator",
        name="{circuit.__name__}_level{level}_valid",
    )
    def test_simulator(self, circuit, level):
        """All the levels with a simulator backend"""
        backend = BasicSimulator()
        result = transpile(circuit(), backend=backend, optimization_level=level, seed_transpiler=42)
        self.assertIsInstance(result, QuantumCircuit)


@ddt
class Test1QV2Working(QiskitTestCase):
    """1QV2 tests that should work."""

    @combine(
        circuit=[emptycircuit],
        level=[0, 1, 2, 3],
        dsc="Transpiling {circuit.__name__} at level {level} should work",
        name="{circuit.__name__}_level{level}_valid_v2",
    )
    def test_device(self, circuit, level):
        """All the levels with all the 1Q backendV2"""
        result = transpile(
            circuit(), backend=Fake1QV2, optimization_level=level, seed_transpiler=42
        )
        self.assertIsInstance(result, QuantumCircuit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Check if a property reached a fixed point."""

from copy import deepcopy

from qiskit.transpiler.basepasses import AnalysisPass


class FixedPoint(AnalysisPass):
    """Check if a property reached a fixed point.

    A dummy analysis pass that checks if a property reached a fixed point.
    The result is saved in ``property_set['<property>_fixed_point']``
    as a boolean.
    """

    def __init__(self, property_to_check):
        """FixedPoint initializer.

        Args:
            property_to_check (str): The property to check if a fixed point was reached.
        """
        super().__init__()
        self._property = property_to_check

    def run(self, dag):
        """Run the FixedPoint pass on `dag`."""
        current_value = self.property_set[self._property]
        fixed_point_previous_property = f"_fixed_point_previous_{self._property}"

        if self.property_set[fixed_point_previous_property] is None:
            self.property_set[f"{self._property}_fixed_point"] = False
        else:
            fixed_point_reached = self.property_set[fixed_point_previous_property] == current_value
            self.property_set[f"{self._property}_fixed_point"] = fixed_point_reached

        self.property_set[fixed_point_previous_property] = deepcopy(current_value)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test of generated fake backends."""
import math
import unittest

from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit, transpile
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.utils import optionals
from test import QiskitTestCase  # pylint: disable=wrong-import-order


def get_test_circuit():
    """Generates simple circuit for tests."""
    desired_vector = [1 / math.sqrt(2), 0, 0, 1 / math.sqrt(2)]
    qr = QuantumRegister(2, "qr")
    cr = ClassicalRegister(2, "cr")
    qc = QuantumCircuit(qr, cr)
    qc.initialize(desired_vector, [qr[0], qr[1]])
    qc.measure(qr[0], cr[0])
    qc.measure(qr[1], cr[1])
    return qc


class FakeBackendsTest(QiskitTestCase):
    """fake backends test."""

    @unittest.skipUnless(optionals.HAS_AER, "qiskit-aer is required to run this test")
    def test_fake_backends_get_kwargs_v1(self):
        """Fake backends honor kwargs passed."""
        backend = GenericBackendV2(num_qubits=5, seed=42)

        qc = QuantumCircuit(2)
        qc.x(range(0, 2))
        qc.measure_all()

        trans_qc = transpile(qc, backend)
        raw_counts = backend.run(trans_qc, shots=1000).result().get_counts()

        self.assertEqual(sum(raw_counts.values()), 1000)

    @unittest.skipUnless(optionals.HAS_AER, "qiskit-aer is required to run this test")
    def test_fake_backend_v2_noise_model_always_present(self):
        """Test that FakeBackendV2 instances always run with noise."""
        backend = GenericBackendV2(num_qubits=5, seed=42)
        backend.set_options(seed_simulator=42)
        qc = QuantumCircuit(1)
        qc.x(0)
        qc.measure_all()

        res = backend.run(qc, shots=1000).result().get_counts()
        # Assert noise was present and result wasn't ideal
        self.assertNotEqual(res, {"1": 1000})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for utilities of Primitives."""

from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
from qiskit.primitives.utils import final_measurement_mapping
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestMapping(QiskitTestCase):
    """Test final_measurement_mapping

    Source:
        https://github.com/Qiskit-Partners/mthree/blob/03ea16fa0f06a9b28e0a19450d88a49501e2c276/mthree/test/test_meas_mapping.py
    """

    def test_empty_circ(self):
        """Empty circuit has no mapping"""
        qc = QuantumCircuit()
        with self.assertWarns(DeprecationWarning):
            self.assertDictEqual(final_measurement_mapping(qc), {})

    def test_sime_circ(self):
        """Just measures"""
        qc = QuantumCircuit(5)
        qc.measure_all()
        with self.assertWarns(DeprecationWarning):
            self.assertDictEqual(final_measurement_mapping(qc), {0: 0, 1: 1, 2: 2, 3: 3, 4: 4})

    def test_simple2_circ(self):
        """Meas followed by Hadamards"""
        qc = QuantumCircuit(5)
        qc.measure_all()
        qc.h(range(5))
        with self.assertWarns(DeprecationWarning):
            self.assertDictEqual(final_measurement_mapping(qc), {})

    def test_multi_qreg(self):
        """Test multiple qregs"""
        qr1 = QuantumRegister(2, "q1")
        qr2 = QuantumRegister(3, "q2")
        cr = ClassicalRegister(5)
        qc = QuantumCircuit(qr1, qr2, cr)

        qc.h(range(5))
        qc.measure(0, 0)
        qc.h(range(5))
        qc.measure(range(2, 4), range(2, 4))
        qc.barrier(range(5))
        qc.measure(1, 4)
        with self.assertWarns(DeprecationWarning):
            self.assertDictEqual(final_measurement_mapping(qc), {2: 2, 3: 3, 1: 4})

    def test_multi_creg(self):
        """Test multiple qregs"""
        qr1 = QuantumRegister(2, "q1")
        qr2 = QuantumRegister(3, "q2")
        cr1 = ClassicalRegister(3, "c1")
        cr2 = ClassicalRegister(2, "c2")
        qc = QuantumCircuit(qr1, qr2, cr1, cr2)

        qc.h(range(5))
        qc.measure(0, 0)
        qc.h(range(5))
        qc.measure(range(2, 4), range(2, 4))
        qc.barrier(range(5))
        qc.measure(1, 4)
        with self.assertWarns(DeprecationWarning):
            self.assertDictEqual(final_measurement_mapping(qc), {2: 2, 3: 3, 1: 4})

    def test_mapping_w_delays(self):
        """Check that measurements followed by delays get in the mapping"""
        qc = QuantumCircuit(2, 2)
        qc.measure(0, 1)
        qc.delay(10, 0)
        qc.measure(1, 0)
        qc.barrier()

        with self.assertWarns(DeprecationWarning):
            maps = final_measurement_mapping(qc)
        self.assertDictEqual(maps, {1: 0, 0: 1})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Standard gates
"""

from .h import HGate, CHGate
from .i import IGate
from .p import PhaseGate, CPhaseGate, MCPhaseGate
from .r import RGate
from .rx import RXGate, CRXGate
from .rxx import RXXGate
from .ry import RYGate, CRYGate
from .ryy import RYYGate
from .rz import RZGate, CRZGate
from .rzz import RZZGate
from .rzx import RZXGate
from .xx_minus_yy import XXMinusYYGate
from .xx_plus_yy import XXPlusYYGate
from .ecr import ECRGate
from .s import SGate, SdgGate, CSGate, CSdgGate
from .swap import SwapGate, CSwapGate
from .iswap import iSwapGate
from .sx import SXGate, SXdgGate, CSXGate
from .dcx import DCXGate
from .t import TGate, TdgGate
from .u import UGate, CUGate
from .u1 import U1Gate, CU1Gate, MCU1Gate
from .u2 import U2Gate
from .u3 import U3Gate, CU3Gate
from .x import XGate, CXGate, CCXGate, C3XGate, C3SXGate, C4XGate, RCCXGate, RC3XGate
from .x import MCXGate, MCXGrayCode, MCXRecursive, MCXVChain
from .y import YGate, CYGate
from .z import ZGate, CZGate, CCZGate
from .global_phase import GlobalPhaseGate
from .multi_control_rotation_gates import mcrx, mcry, mcrz


def get_standard_gate_name_mapping():
    """Return a dictionary mapping the name of standard gates and instructions to an object for
    that name.

    Examples:

        .. code-block:: python

            from qiskit.circuit.library import get_standard_gate_name_mapping

            gate_name_map = get_standard_gate_name_mapping()
            cx_object = gate_name_map["cx"]

            print(cx_object)
            print(type(cx_object))

        .. code-block:: text

            Instruction(name='cx', num_qubits=2, num_clbits=0, params=[])
            _SingletonCXGate
    """

    from qiskit.circuit.parameter import Parameter
    from qiskit.circuit.measure import Measure
    from qiskit.circuit.delay import Delay
    from qiskit.circuit.reset import Reset

    lambda_ = Parameter("λ")
    theta = Parameter("ϴ")
    phi = Parameter("φ")
    gamma = Parameter("γ")
    beta = Parameter("β")
    time = Parameter("t")

    # Standard gates library mapping, multicontrolled gates not included since they're
    # variable width
    gates = [
        IGate(),
        SXGate(),
        XGate(),
        CXGate(),
        RZGate(lambda_),
        RGate(theta, phi),
        C3SXGate(),
        CCXGate(),
        DCXGate(),
        CHGate(),
        CPhaseGate(theta),
        CRXGate(theta),
        CRYGate(theta),
        CRZGate(theta),
        CSwapGate(),
        CSXGate(),
        CUGate(theta, phi, lambda_, gamma),
        CU1Gate(lambda_),
        CU3Gate(theta, phi, lambda_),
        CYGate(),
        CZGate(),
        CCZGate(),
        GlobalPhaseGate(theta),
        HGate(),
        PhaseGate(theta),
        RCCXGate(),
        RC3XGate(),
        RXGate(theta),
        RXXGate(theta),
        RYGate(theta),
        RYYGate(theta),
        RZZGate(theta),
        RZXGate(theta),
        XXMinusYYGate(theta, beta),
        XXPlusYYGate(theta, beta),
        ECRGate(),
        SGate(),
        SdgGate(),
        CSGate(),
        CSdgGate(),
        SwapGate(),
        iSwapGate(),
        SXdgGate(),
        TGate(),
        TdgGate(),
        UGate(theta, phi, lambda_),
        U1Gate(lambda_),
        U2Gate(phi, lambda_),
        U3Gate(theta, phi, lambda_),
        YGate(),
        ZGate(),
        Delay(time),
        Reset(),
        Measure(),
    ]
    name_mapping = {gate.name: gate for gate in gates}
    return name_mapping



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Diagonal matrix circuit."""

from __future__ import annotations
from collections.abc import Sequence

import cmath
import math
import numpy as np

from qiskit.circuit.gate import Gate
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit.annotated_operation import AnnotatedOperation, InverseModifier
from qiskit.utils.deprecation import deprecate_func

from .ucrz import UCRZGate

_EPS = 1e-10


class Diagonal(QuantumCircuit):
    """Circuit implementing a diagonal transformation."""

    @deprecate_func(since="1.3", additional_msg="Use DiagonalGate instead.", pending=True)
    def __init__(self, diag: Sequence[complex]) -> None:
        r"""
        Args:
            diag: List of the :math:`2^k` diagonal entries (for a diagonal gate on :math:`k` qubits).

        Raises:
            CircuitError: if the list of the diagonal entries or the qubit list is in bad format;
                if the number of diagonal entries is not :math:`2^k`, where :math:`k` denotes the
                number of qubits.
        """
        DiagonalGate._check_input(diag)
        num_qubits = int(math.log2(len(diag)))

        super().__init__(num_qubits, name="Diagonal")
        self.append(DiagonalGate(diag), self.qubits)


class DiagonalGate(Gate):
    r"""A generic diagonal quantum gate.

    Matrix form:

    .. math::
        \text{DiagonalGate}\ q_0, q_1, .., q_{n-1} =
            \begin{pmatrix}
                D[0]    & 0         & \dots     & 0 \\
                0       & D[1]      & \dots     & 0 \\
                \vdots  & \vdots    & \ddots    & 0 \\
                0       & 0         & \dots     & D[n-1]
            \end{pmatrix}

    Diagonal gates are useful as representations of Boolean functions,
    as they can map from :math:`\{0,1\}^{2^n}` to :math:`\{0,1\}^{2^n}` space. For example a phase
    oracle can be seen as a diagonal gate with :math:`\{1, -1\}` on the diagonals. Such
    an oracle will induce a :math:`+1` or :math`-1` phase on the amplitude of any corresponding
    basis state.

    Diagonal gates appear in many classically hard oracular problems such as
    Forrelation or Hidden Shift circuits.

    Diagonal gates are represented and simulated more efficiently than a dense
    :math:`2^n \times 2^n` unitary matrix.

    The reference implementation is via the method described in
    Theorem 7 of [1]. The code is based on Emanuel Malvetti's semester thesis
    at ETH in 2018, supervised by Raban Iten and Prof. Renato Renner.

    **Reference:**

    [1] Shende et al., Synthesis of Quantum Logic Circuits, 2009
    `arXiv:0406176 <https://arxiv.org/pdf/quant-ph/0406176.pdf>`_
    """

    def __init__(self, diag: Sequence[complex]) -> None:
        r"""
        Args:
            diag: list of the :math:`2^k` diagonal entries (for a diagonal gate on :math:`k` qubits).
        """
        self._check_input(diag)
        num_qubits = int(math.log2(len(diag)))

        super().__init__("diagonal", num_qubits, diag)

    def _define(self):
        # Since the diagonal is a unitary, all its entries have absolute value
        # one and the diagonal is fully specified by the phases of its entries.
        diag_phases = [cmath.phase(z) for z in self.params]
        n = len(diag_phases)
        circuit = QuantumCircuit(self.num_qubits)

        while n >= 2:
            angles_rz = []
            for i in range(0, n, 2):
                diag_phases[i // 2], rz_angle = _extract_rz(diag_phases[i], diag_phases[i + 1])
                angles_rz.append(rz_angle)
            num_act_qubits = int(math.log2(n))
            ctrl_qubits = list(range(self.num_qubits - num_act_qubits + 1, self.num_qubits))
            target_qubit = self.num_qubits - num_act_qubits

            ucrz = UCRZGate(angles_rz)
            circuit.append(ucrz, [target_qubit] + ctrl_qubits)

            n //= 2
        circuit.global_phase += diag_phases[0]

        self.definition = circuit

    def validate_parameter(self, parameter):
        """Diagonal Gate parameter should accept complex
        (in addition to the Gate parameter types) and always return build-in complex."""
        if isinstance(parameter, complex):
            return complex(parameter)
        else:
            return complex(super().validate_parameter(parameter))

    def inverse(self, annotated: bool = False):
        """Return the inverse of the diagonal gate."""
        if annotated:
            return AnnotatedOperation(self.copy(), InverseModifier)

        return DiagonalGate([np.conj(entry) for entry in self.params])

    @staticmethod
    def _check_input(diag):
        """Check if ``diag`` is in valid format."""
        if not isinstance(diag, (list, np.ndarray)):
            raise CircuitError("Diagonal entries must be in a list or numpy array.")
        num_qubits = math.log2(len(diag))
        if num_qubits < 1 or not num_qubits.is_integer():
            raise CircuitError("The number of diagonal entries is not a positive power of 2.")
        if not np.allclose(np.abs(diag), 1, atol=_EPS):
            raise CircuitError("A diagonal element does not have absolute value one.")


def _extract_rz(phi1, phi2):
    """
    Extract a Rz rotation (angle given by first output) such that exp(j*phase)*Rz(z_angle)
    is equal to the diagonal matrix with entires exp(1j*ph1) and exp(1j*ph2).
    """
    phase = (phi1 + phi2) / 2.0
    z_angle = phi2 - phi1
    return phase, z_angle



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Choose a Layout by finding the most connected subset of qubits."""


import numpy as np
import rustworkx

from qiskit.transpiler.layout import Layout
from qiskit.transpiler.basepasses import AnalysisPass
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.passes.layout import disjoint_utils

from qiskit.utils import deprecate_arg

from qiskit._accelerate.dense_layout import best_subset


class DenseLayout(AnalysisPass):
    """Choose a Layout by finding the most connected subset of qubits.

    This pass associates a physical qubit (int) to each virtual qubit
    of the circuit (Qubit).

    Note:
        Even though a ``'layout'`` is not strictly a property of the DAG,
        in the transpiler architecture it is best passed around between passes
        by being set in ``property_set``.
    """

    @deprecate_arg(
        name="backend_prop",
        since="1.4",
        package_name="Qiskit",
        removal_timeline="in Qiskit 2.0",
        additional_msg="The BackendProperties data structure has been deprecated and will be "
        "removed in Qiskit 2.0. The `target` input argument should be used instead. "
        "You can use Target.from_configuration() to build the target from the properties "
        "object, but in 2.0 you will need to generate a target directly.",
    )
    def __init__(self, coupling_map=None, backend_prop=None, target=None):
        """DenseLayout initializer.

        Args:
            coupling_map (Coupling): directed graph representing a coupling map.
            backend_prop (BackendProperties): backend properties object
            target (Target): A target representing the target backend.
        """
        super().__init__()
        self.coupling_map = coupling_map
        self.backend_prop = backend_prop
        self.target = target
        self.adjacency_matrix = None
        if target is not None:
            self.coupling_map = target.build_coupling_map()

    def run(self, dag):
        """Run the DenseLayout pass on `dag`.

        Pick a convenient layout depending on the best matching
        qubit connectivity, and set the property `layout`.

        Args:
            dag (DAGCircuit): DAG to find layout for.

        Raises:
            TranspilerError: if dag wider than self.coupling_map
        """
        if self.coupling_map is None:
            raise TranspilerError(
                "A coupling_map or target with constrained qargs is necessary to run the pass."
            )
        layout_components = disjoint_utils.run_pass_over_connected_components(
            dag,
            self.coupling_map if self.target is None else self.target,
            self._inner_run,
        )
        layout_mapping = {}
        for component in layout_components:
            layout_mapping.update(component)
        layout = Layout(layout_mapping)
        for qreg in dag.qregs.values():
            layout.add_register(qreg)
        self.property_set["layout"] = layout

    def _inner_run(self, dag, coupling_map):
        num_dag_qubits = len(dag.qubits)
        if num_dag_qubits > coupling_map.size():
            raise TranspilerError("Number of qubits greater than device.")
        num_cx = 0
        num_meas = 0

        if self.target is not None:
            num_cx = 1
            num_meas = 1
        else:
            # Get avg number of cx and meas per qubit
            ops = dag.count_ops(recurse=True)
            if "cx" in ops.keys():
                num_cx = ops["cx"]
            if "measure" in ops.keys():
                num_meas = ops["measure"]

        best_sub = self._best_subset(num_dag_qubits, num_meas, num_cx, coupling_map)
        layout_mapping = {
            qubit: coupling_map.graph[int(best_sub[i])] for i, qubit in enumerate(dag.qubits)
        }
        return layout_mapping

    def _best_subset(self, num_qubits, num_meas, num_cx, coupling_map):
        """Computes the qubit mapping with the best connectivity.

        Args:
            num_qubits (int): Number of subset qubits to consider.

        Returns:
            ndarray: Array of qubits to use for best connectivity mapping.
        """
        from scipy.sparse import coo_matrix, csgraph

        if num_qubits == 1:
            return np.array([0])
        if num_qubits == 0:
            return []

        adjacency_matrix = rustworkx.adjacency_matrix(coupling_map.graph)
        reverse_index_map = {v: k for k, v in enumerate(coupling_map.graph.nodes())}

        error_mat, use_error = _build_error_matrix(
            coupling_map.size(),
            reverse_index_map,
            backend_prop=self.backend_prop,
            coupling_map=self.coupling_map,
            target=self.target,
        )

        rows, cols, best_map = best_subset(
            num_qubits,
            adjacency_matrix,
            num_meas,
            num_cx,
            use_error,
            coupling_map.is_symmetric,
            error_mat,
        )
        data = [1] * len(rows)
        sp_sub_graph = coo_matrix((data, (rows, cols)), shape=(num_qubits, num_qubits)).tocsr()
        perm = csgraph.reverse_cuthill_mckee(sp_sub_graph)
        best_map = best_map[perm]
        return best_map


def _build_error_matrix(num_qubits, qubit_map, target=None, coupling_map=None, backend_prop=None):
    error_mat = np.zeros((num_qubits, num_qubits))
    use_error = False
    if target is not None and target.qargs is not None:
        for qargs in target.qargs:
            # Ignore gates over 2q DenseLayout only works with 2q
            if len(qargs) > 2:
                continue
            error = 0.0
            ops = target.operation_names_for_qargs(qargs)
            for op in ops:
                props = target[op].get(qargs, None)
                if props is not None and props.error is not None:
                    # Use max error rate to represent operation error
                    # on a qubit(s). If there is more than 1 operation available
                    # we don't know what will be used on the qubits eventually
                    # so we take the highest error operation as a proxy for
                    # the possible worst case.
                    error = max(error, props.error)
            max_error = error
            if any(qubit not in qubit_map for qubit in qargs):
                continue
            # TODO: Factor in T1 and T2 to error matrix after #7736
            if len(qargs) == 1:
                qubit = qubit_map[qargs[0]]
                error_mat[qubit][qubit] = max_error
                use_error = True
            elif len(qargs) == 2:
                error_mat[qubit_map[qargs[0]]][qubit_map[qargs[1]]] = max_error
                use_error = True
    elif backend_prop and coupling_map:
        error_dict = {
            tuple(gate.qubits): gate.parameters[0].value
            for gate in backend_prop.gates
            if len(gate.qubits) == 2
        }
        for edge in coupling_map.get_edges():
            gate_error = error_dict.get(edge)
            if gate_error is not None:
                if edge[0] not in qubit_map or edge[1] not in qubit_map:
                    continue
                error_mat[qubit_map[edge[0]]][qubit_map[edge[1]]] = gate_error
                use_error = True
        for index, qubit_data in enumerate(backend_prop.qubits):
            if index not in qubit_map:
                continue
            for item in qubit_data:
                if item.name == "readout_error":
                    mapped_index = qubit_map[index]
                    error_mat[mapped_index][mapped_index] = item.value
                    use_error = True
    return error_mat, use_error



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=unsubscriptable-object

"""Test Qiskit's Instruction class."""

import unittest.mock
import numpy as np

from qiskit.circuit import (
    Gate,
    Parameter,
    Instruction,
    InstructionSet,
    QuantumCircuit,
    QuantumRegister,
    ClassicalRegister,
    Qubit,
    Clbit,
    IfElseOp,
)
from qiskit.circuit.library import HGate, RZGate, CXGate, SGate, SdgGate, TGate
from qiskit.circuit.annotated_operation import AnnotatedOperation
from qiskit.circuit.classical import expr
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit.random import random_circuit
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestInstructions(QiskitTestCase):
    """Instructions tests."""

    def test_instructions_equal(self):
        """Test equality of two instructions."""
        hop1 = Instruction("h", 1, 0, [])
        hop2 = Instruction("s", 1, 0, [])
        hop3 = Instruction("h", 1, 0, [])

        self.assertFalse(hop1 == hop2)
        self.assertTrue(hop1 == hop3)

        uop1 = Instruction("u", 1, 0, [0.4, 0.5, 0.5])
        uop2 = Instruction("u", 1, 0, [0.4, 0.6, 0.5])
        uop3 = Instruction("v", 1, 0, [0.4, 0.5, 0.5])
        uop4 = Instruction("u", 1, 0, [0.4, 0.5, 0.5])

        self.assertFalse(uop1 == uop2)
        self.assertTrue(uop1 == uop4)
        self.assertFalse(uop1 == uop3)

        self.assertTrue(HGate() == HGate())
        self.assertFalse(HGate() == CXGate())
        self.assertFalse(hop1 == HGate())

        eop1 = Instruction("kraus", 1, 0, [np.array([[1, 0], [0, 1]])])
        eop2 = Instruction("kraus", 1, 0, [np.array([[0, 1], [1, 0]])])
        eop3 = Instruction("kraus", 1, 0, [np.array([[1, 0], [0, 1]])])
        eop4 = Instruction("kraus", 1, 0, [np.eye(4)])

        self.assertTrue(eop1 == eop3)
        self.assertFalse(eop1 == eop2)
        self.assertFalse(eop1 == eop4)

    def test_instructions_equal_with_parameters(self):
        """Test equality of instructions for cases with Parameters."""
        theta = Parameter("theta")
        phi = Parameter("phi")

        # Verify we can check params including parameters
        self.assertEqual(
            Instruction("u", 1, 0, [theta, phi, 0.4]), Instruction("u", 1, 0, [theta, phi, 0.4])
        )

        # Verify we can test for correct parameter order
        self.assertNotEqual(
            Instruction("u", 1, 0, [theta, phi, 0]), Instruction("u", 1, 0, [phi, theta, 0])
        )

        # Verify we can still find a wrong fixed param if we use parameters
        self.assertNotEqual(
            Instruction("u", 1, 0, [theta, phi, 0.4]), Instruction("u", 1, 0, [theta, phi, 0.5])
        )

        # Verify we can find cases when param != float
        self.assertNotEqual(
            Instruction("u", 1, 0, [0.3, phi, 0.4]), Instruction("u", 1, 0, [theta, phi, 0.5])
        )

    def test_instructions_soft_compare(self):
        """Test soft comparison between instructions."""
        theta = Parameter("theta")
        phi = Parameter("phi")

        # Verify that we are insensitive when there are parameters.
        self.assertTrue(
            Instruction("u", 1, 0, [0.3, phi, 0.4]).soft_compare(
                Instruction("u", 1, 0, [theta, phi, 0.4])
            )
        )

        # Verify that normal equality still holds.
        self.assertTrue(
            Instruction("u", 1, 0, [0.4, 0.5]).soft_compare(Instruction("u", 1, 0, [0.4, 0.5]))
        )

        # Test that when names differ we get False.
        self.assertFalse(
            Instruction("u", 1, 0, [0.4, phi]).soft_compare(Instruction("v", 1, 0, [theta, phi]))
        )

        # Test that when names are the same but number of qubits differ we get False
        self.assertFalse(Instruction("u", 1, 0, []).soft_compare(Instruction("u", 2, 0, [])))

        # Test that when names are the same but number of clbits differ we get False
        self.assertFalse(Instruction("u", 1, 0, []).soft_compare(Instruction("u", 1, 1, [])))

        # Test cutoff precision.
        self.assertFalse(
            Instruction("v", 1, 0, [0.401, phi]).soft_compare(Instruction("v", 1, 0, [0.4, phi]))
        )

        # Test cutoff precision.
        self.assertTrue(
            Instruction("v", 1, 0, [0.4 + 1.0e-20, phi]).soft_compare(
                Instruction("v", 1, 0, [0.4, phi])
            )
        )

    def test_instructions_equal_with_parameter_expressions(self):
        """Test equality of instructions for cases with ParameterExpressions."""
        theta = Parameter("theta")
        phi = Parameter("phi")
        sum_ = theta + phi
        product_ = theta * phi

        # Verify we can check params including parameters
        self.assertEqual(
            Instruction("u", 1, 0, [sum_, product_, 0.4]),
            Instruction("u", 1, 0, [sum_, product_, 0.4]),
        )

        # Verify we can test for correct parameter order
        self.assertNotEqual(
            Instruction("u", 1, 0, [product_, sum_, 0]), Instruction("u", 1, 0, [sum_, product_, 0])
        )

        # Verify we can still find a wrong fixed param if we use parameters
        self.assertNotEqual(
            Instruction("u", 1, 0, [sum_, phi, 0.4]), Instruction("u", 1, 0, [sum_, phi, 0.5])
        )

        # Verify we can find cases when param != float
        self.assertNotEqual(
            Instruction("u", 1, 0, [0.3, sum_, 0.4]), Instruction("u", 1, 0, [product_, sum_, 0.5])
        )

    def circuit_instruction_circuit_roundtrip(self):
        """test converting between circuit and instruction and back
        preserves the circuit"""
        q = QuantumRegister(4)
        c = ClassicalRegister(4)
        circ1 = QuantumCircuit(q, c, name="circuit1")
        circ1.h(q[0])
        circ1.crz(0.1, q[0], q[1])
        circ1.id(q[1])
        circ1.u(0.1, 0.2, -0.2, q[0])
        circ1.barrier()
        circ1.measure(q, c)
        with self.assertWarns(DeprecationWarning):
            circ1.rz(0.8, q[0]).c_if(c, 6)
        inst = circ1.to_instruction()

        circ2 = QuantumCircuit(q, c, name="circ2")
        circ2.append(inst, q[:])

        self.assertEqual(circ1, circ2)

    def test_append_opaque_wrong_dimension(self):
        """test appending opaque gate to wrong dimension wires."""
        qr = QuantumRegister(2)
        circ = QuantumCircuit(qr)
        opaque_gate = Gate(name="crz_2", num_qubits=2, params=[0.5])
        self.assertRaises(CircuitError, circ.append, opaque_gate, [qr[0]])

    def test_opaque_gate(self):
        """test opaque gate functionality"""
        q = QuantumRegister(4)
        c = ClassicalRegister(4)
        circ = QuantumCircuit(q, c, name="circ")
        opaque_gate = Gate(name="crz_2", num_qubits=2, params=[0.5])
        circ.append(opaque_gate, [q[2], q[0]])
        self.assertEqual(circ.data[0].operation.name, "crz_2")
        self.assertEqual(circ.decompose(), circ)

    def test_opaque_instruction(self):
        """test opaque instruction does not decompose"""
        q = QuantumRegister(4)
        c = ClassicalRegister(2)
        circ = QuantumCircuit(q, c)
        opaque_inst = Instruction(name="my_inst", num_qubits=3, num_clbits=1, params=[0.5])
        circ.append(opaque_inst, [q[3], q[1], q[0]], [c[1]])
        self.assertEqual(circ.data[0].operation.name, "my_inst")
        self.assertEqual(circ.decompose(), circ)

    def test_reverse_gate(self):
        """test reversing a composite gate"""
        q = QuantumRegister(4)
        circ = QuantumCircuit(q, name="circ")
        circ.h(q[0])
        circ.crz(0.1, q[0], q[1])
        circ.id(q[1])
        circ.u(0.1, 0.2, -0.2, q[0])
        gate = circ.to_gate()

        circ = QuantumCircuit(q, name="circ")
        circ.u(0.1, 0.2, -0.2, q[0])
        circ.id(q[1])
        circ.crz(0.1, q[0], q[1])
        circ.h(q[0])
        gate_reverse = circ.to_gate()
        self.assertEqual(gate.reverse_ops().definition, gate_reverse.definition)

    def test_reverse_instruction(self):
        """test reverseing an instruction with conditionals"""
        q = QuantumRegister(4)
        c = ClassicalRegister(4)
        circ = QuantumCircuit(q, c, name="circ")
        circ.t(q[1])
        circ.u(0.1, 0.2, -0.2, q[0])
        circ.barrier()
        circ.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            circ.rz(0.8, q[0]).c_if(c, 6)
        with self.assertWarns(DeprecationWarning):
            inst = circ.to_instruction()

        circ = QuantumCircuit(q, c, name="circ")
        with self.assertWarns(DeprecationWarning):
            circ.rz(0.8, q[0]).c_if(c, 6)
        circ.measure(q[0], c[0])
        circ.barrier()
        circ.u(0.1, 0.2, -0.2, q[0])
        circ.t(q[1])
        with self.assertWarns(DeprecationWarning):
            inst_reverse = circ.to_instruction()

        self.assertEqual(inst.reverse_ops().definition, inst_reverse.definition)

    def test_reverse_opaque(self):
        """test opaque gates reverse to themselves"""
        opaque_gate = Gate(name="crz_2", num_qubits=2, params=[0.5])
        self.assertEqual(opaque_gate.reverse_ops(), opaque_gate)
        hgate = HGate()
        self.assertEqual(hgate.reverse_ops(), hgate)

    def test_inverse_and_append(self):
        """test appending inverted gates to circuits"""
        q = QuantumRegister(1)
        circ = QuantumCircuit(q, name="circ")
        circ.s(q)
        circ.append(SGate().inverse(), q[:])
        circ.append(TGate().inverse(), q[:])
        circ.t(q)
        gate = circ.to_instruction()
        circ = QuantumCircuit(q, name="circ")
        circ.inverse()
        circ.tdg(q)
        circ.t(q)
        circ.s(q)
        circ.sdg(q)
        gate_inverse = circ.to_instruction()
        self.assertEqual(gate.inverse().definition, gate_inverse.definition)

    def test_inverse_composite_gate(self):
        """test inverse of composite gate"""
        q = QuantumRegister(4)
        circ = QuantumCircuit(q, name="circ")
        circ.h(q[0])
        circ.crz(0.1, q[0], q[1])
        circ.id(q[1])
        circ.u(0.1, 0.2, -0.2, q[0])
        gate = circ.to_instruction()
        circ = QuantumCircuit(q, name="circ")
        circ.u(-0.1, 0.2, -0.2, q[0])
        circ.id(q[1])
        circ.crz(-0.1, q[0], q[1])
        circ.h(q[0])
        gate_inverse = circ.to_instruction()
        self.assertEqual(gate.inverse().definition, gate_inverse.definition)

    def test_inverse_recursive(self):
        """test that a hierarchical gate recursively inverts"""
        qr0 = QuantumRegister(2)
        circ0 = QuantumCircuit(qr0, name="circ0")
        circ0.t(qr0[0])
        circ0.rx(0.4, qr0[1])
        circ0.cx(qr0[1], qr0[0])
        little_gate = circ0.to_instruction()

        qr1 = QuantumRegister(4)
        circ1 = QuantumCircuit(qr1, name="circuit1")
        circ1.cp(-0.1, qr1[0], qr1[2])
        circ1.id(qr1[1])
        circ1.append(little_gate, [qr1[2], qr1[3]])

        circ_inv = QuantumCircuit(qr1, name="circ1_dg")
        circ_inv.append(little_gate.inverse(), [qr1[2], qr1[3]])
        circ_inv.id(qr1[1])
        circ_inv.cp(0.1, qr1[0], qr1[2])

        self.assertEqual(circ1.inverse(), circ_inv)

    def test_inverse_instruction_with_measure(self):
        """test inverting instruction with measure fails"""
        q = QuantumRegister(4)
        c = ClassicalRegister(4)
        circ = QuantumCircuit(q, c, name="circ")
        circ.t(q[1])
        circ.u(0.1, 0.2, -0.2, q[0])
        circ.barrier()
        circ.measure(q[0], c[0])
        inst = circ.to_instruction()
        self.assertRaises(CircuitError, inst.inverse)

    def test_inverse_instruction_with_conditional(self):
        """test inverting instruction with conditionals fails"""
        q = QuantumRegister(4)
        c = ClassicalRegister(4)
        circ = QuantumCircuit(q, c, name="circ")
        circ.t(q[1])
        circ.u(0.1, 0.2, -0.2, q[0])
        circ.barrier()
        circ.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            circ.rz(0.8, q[0]).c_if(c, 6)
        with self.assertWarns(DeprecationWarning):
            inst = circ.to_instruction()
        self.assertRaises(CircuitError, inst.inverse)

    def test_inverse_opaque(self):
        """test inverting opaque gate fails"""
        opaque_gate = Gate(name="crz_2", num_qubits=2, params=[0.5])
        self.assertRaises(CircuitError, opaque_gate.inverse)

    def test_inverse_empty(self):
        """test inverting empty gate works"""
        q = QuantumRegister(3)
        c = ClassicalRegister(3)
        empty_circ = QuantumCircuit(q, c, name="empty_circ")
        empty_gate = empty_circ.to_instruction()
        self.assertEqual(empty_gate.inverse().definition, empty_gate.definition)

    def test_inverse_with_global_phase(self):
        """test inverting instruction with global phase in definition."""
        q = QuantumRegister(1)
        circ = QuantumCircuit(q, name="circ", global_phase=np.pi / 3)
        circ.x(q)
        gate = circ.to_instruction()
        circ = QuantumCircuit(q, name="circ", global_phase=-np.pi / 3)
        circ.x(q)
        gate_inverse = circ.to_instruction()
        self.assertEqual(gate.inverse().definition, gate_inverse.definition)

    def test_inverse_with_label(self):
        """test inverting gate initialized with label attribute."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q, name="circ")
        qc.cx(0, 1)
        qc_gate = qc.to_gate()
        qc_gate_inverse = qc_gate.inverse()
        self.assertEqual(qc_gate.name + "_dg", qc_gate_inverse.name)
        qc_gate_inverse_inverse = qc_gate_inverse.inverse()
        self.assertEqual(qc_gate_inverse_inverse.name, qc_gate.name)

    def test_no_broadcast(self):
        """See https://github.com/Qiskit/qiskit-terra/issues/2777
        When creating custom instructions, do not broadcast parameters"""
        qr = QuantumRegister(2)
        cr = ClassicalRegister(2)
        subcircuit = QuantumCircuit(qr, cr, name="subcircuit")

        subcircuit.x(qr[0])
        subcircuit.h(qr[1])
        subcircuit.measure(qr[0], cr[0])
        subcircuit.measure(qr[1], cr[1])

        inst = subcircuit.to_instruction()
        circuit = QuantumCircuit(qr, cr, name="circuit")
        circuit.append(inst, qr[:], cr[:])
        self.assertEqual(circuit.qregs, [qr])
        self.assertEqual(circuit.cregs, [cr])
        self.assertEqual(circuit.qubits, [qr[0], qr[1]])
        self.assertEqual(circuit.clbits, [cr[0], cr[1]])

    def test_modifying_copied_params_leaves_orig(self):
        """Verify modifying the parameters of a copied instruction does not
        affect the original."""

        inst = Instruction("test", 2, 1, [0, 1, 2])

        cpy = inst.copy()

        cpy.params[1] = 7

        self.assertEqual(inst.params, [0, 1, 2])

    def test_instance_of_instruction(self):
        """Test correct error message is raised when invalid instruction
        is passed to append"""

        qr = QuantumRegister(2)
        qc = QuantumCircuit(qr)
        with self.assertRaisesRegex(CircuitError, r"Object is a subclass of Operation"):
            qc.append(HGate, qr[:], [])

    def test_repr_of_instructions(self):
        """Test the __repr__ method of the Instruction
        class"""

        ins1 = Instruction("test_instruction", 3, 5, [0, 1, 2, 3])
        self.assertEqual(
            repr(ins1),
            f"Instruction(name='{ins1.name}', num_qubits={ins1.num_qubits}, "
            f"num_clbits={ins1.num_clbits}, params={ins1.params})",
        )

        ins2 = random_circuit(num_qubits=4, depth=4, measure=True).to_instruction()
        self.assertEqual(
            repr(ins2),
            f"Instruction(name='{ins2.name}', num_qubits={ins2.num_qubits}, "
            f"num_clbits={ins2.num_clbits}, params={ins2.params})",
        )

    def test_instruction_condition_bits(self):
        """Test that the ``condition_bits`` property behaves correctly until it is deprecated and
        removed."""
        bits = [Clbit(), Clbit()]
        cr1 = ClassicalRegister(2, "cr1")
        cr2 = ClassicalRegister(2, "cr2")
        body = QuantumCircuit(cr1, cr2, bits)

        def key(bit):
            return body.find_bit(bit).index

        op = IfElseOp((bits[0], False), body)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(op.condition_bits, [bits[0]])

        op = IfElseOp((cr1, 3), body)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(op.condition_bits, list(cr1))

        op = IfElseOp(expr.logic_and(bits[1], expr.equal(cr2, 3)), body)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(
                sorted(op.condition_bits, key=key), sorted([bits[1]] + list(cr2), key=key)
            )

    def test_instructionset_c_if_direct_resource(self):
        """Test that using :meth:`.InstructionSet.c_if` with an exact classical resource always
        works, and produces the expected condition."""
        cr1 = ClassicalRegister(3)
        qubits = [Qubit()]
        loose_clbits = [Clbit(), Clbit(), Clbit()]
        # These bits are going into registers which overlap.
        register_clbits = [Clbit(), Clbit(), Clbit()]
        cr2 = ClassicalRegister(bits=register_clbits[:2])
        cr3 = ClassicalRegister(bits=register_clbits[1:])

        def case(resource):
            qc = QuantumCircuit(cr1, qubits, loose_clbits, cr2, cr3)
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(resource, 0)
            with self.assertWarns(DeprecationWarning):
                c_if_resource = qc.data[0].operation.condition[0]
            self.assertIs(c_if_resource, resource)

        with self.subTest("classical register"):
            case(cr1)
        with self.subTest("bit from classical register"):
            case(cr1[0])
        with self.subTest("loose bit"):
            case(loose_clbits[0])
        with self.subTest("overlapping register left"):
            case(cr2)
        with self.subTest("overlapping register right"):
            case(cr3)
        with self.subTest("bit in two different registers"):
            case(register_clbits[1])

    def test_instructionset_c_if_indexing(self):
        """Test that using :meth:`.InstructionSet.c_if` with an index for the classical resource
        resolves to the same value that :obj:`.QuantumCircuit` would resolve it to.

        Regression test for gh-7246."""
        cr1 = ClassicalRegister(3)
        qubits = [Qubit()]
        loose_clbits = [Clbit(), Clbit(), Clbit()]
        # These bits are going into registers which overlap.
        register_clbits = [Clbit(), Clbit(), Clbit()]
        cr2 = ClassicalRegister(bits=register_clbits[:2])
        cr3 = ClassicalRegister(bits=register_clbits[1:])

        qc = QuantumCircuit(cr1, qubits, loose_clbits, cr2, cr3)
        for index, clbit in enumerate(qc.clbits):
            with self.subTest(index=index):
                with self.assertWarns(DeprecationWarning):
                    qc.x(0).c_if(index, 0)
                qc.measure(0, index)
                with self.assertWarns(DeprecationWarning):
                    from_c_if = qc.data[-2].operation.condition[0]
                from_measure = qc.data[-1].clbits[0]
                self.assertIs(from_c_if, from_measure)
                # Sanity check that the bit is also the one we expected.
                self.assertIs(from_c_if, clbit)

    def test_instructionset_c_if_size_1_classical_register(self):
        """Test that there is a distinction between conditioning on a single bit and a classical
        register, even if the register in question has a size of one."""
        qr = QuantumRegister(1)
        cr = ClassicalRegister(1)
        qc = QuantumCircuit(qr, cr)

        with self.subTest("classical register"):
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(cr, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(qc.data[-1].operation.condition[0], cr)
        with self.subTest("classical bit by value"):
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(cr[0], 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(qc.data[-1].operation.condition[0], cr[0])
        with self.subTest("classical bit by index"):
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(0, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(qc.data[-1].operation.condition[0], cr[0])

    def test_instructionset_c_if_no_classical_registers(self):
        """Test that using :meth:`.InstructionSet.c_if` works if there are no classical registers
        defined on the circuit.

        Regression test for gh-7250."""
        bits = [Qubit(), Clbit()]
        qc = QuantumCircuit(bits)
        with self.subTest("by value"):
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(bits[1], 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(qc.data[-1].operation.condition[0], bits[1])
        with self.subTest("by index"):
            with self.assertWarns(DeprecationWarning):
                qc.x(0).c_if(0, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(qc.data[-1].operation.condition[0], bits[1])

    def test_instructionset_c_if_rejects_invalid_specifiers(self):
        """Test that calling the :meth:`.InstructionSet.c_if` method on instructions added to a
        circuit raises a suitable exception if an invalid specifier is passed to it."""

        qreg = QuantumRegister(1)
        creg = ClassicalRegister(2)

        def case(specifier, message):
            qc = QuantumCircuit(qreg, creg)
            instruction = qc.x(0)
            with self.assertRaisesRegex(CircuitError, message):
                with self.assertWarns(DeprecationWarning):
                    instruction.c_if(specifier, 0)

        with self.subTest("absent bit"):
            case(Clbit(), r"Clbit .* is not present in this circuit\.")
        with self.subTest("absent register"):
            case(ClassicalRegister(2), r"Register .* is not present in this circuit\.")
        with self.subTest("index out of range"):
            case(2, r"Classical bit index .* is out-of-range\.")
        with self.subTest("list of bits"):
            case(list(creg), r"Unknown classical resource specifier: .*")
        with self.subTest("tuple of bits"):
            case(tuple(creg), r"Unknown classical resource specifier: .*")
        with self.subTest("float"):
            case(1.0, r"Unknown classical resource specifier: .*")

    def test_instructionset_c_if_with_no_requester(self):
        """Test that using a raw :obj:`.InstructionSet` with no classical-resource resolver accepts
        arbitrary :obj:`.Clbit` and `:obj:`.ClassicalRegister` instances, but rejects integers."""

        with self.subTest("accepts arbitrary register"):
            instruction = RZGate(0)
            instructions = InstructionSet()
            instructions.add(instruction, [Qubit()], [])
            register = ClassicalRegister(2)
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(register, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(instructions[0].operation.condition[0], register)
        with self.subTest("accepts arbitrary bit"):
            instruction = RZGate(0)
            instructions = InstructionSet()
            instructions.add(instruction, [Qubit()], [])
            bit = Clbit()
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(bit, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(instructions[0].operation.condition[0], bit)
        with self.subTest("rejects index"):
            instruction = RZGate(0)
            instructions = InstructionSet()
            instructions.add(instruction, [Qubit()], [])
            with self.assertRaisesRegex(CircuitError, r"Cannot pass an index as a condition .*"):
                with self.assertWarns(DeprecationWarning):
                    instructions.c_if(0, 0)

    def test_instructionset_c_if_calls_custom_requester(self):
        """Test that :meth:`.InstructionSet.c_if` calls a custom requester, and uses its output."""
        # This isn't expected to be useful to end users, it's more about the principle that you can
        # control the resolution paths, so future blocking constructs can forbid the method from
        # accessing certain resources.

        sentinel_bit = Clbit()
        sentinel_register = ClassicalRegister(2)

        def dummy_requester(specifier):
            """A dummy requester that returns sentinel values."""
            if not isinstance(specifier, (int, Clbit, ClassicalRegister)):
                raise CircuitError
            return sentinel_bit if isinstance(specifier, (int, Clbit)) else sentinel_register

        dummy_requester = unittest.mock.MagicMock(wraps=dummy_requester)

        with self.subTest("calls requester with bit"):
            dummy_requester.reset_mock()
            instruction = RZGate(0)
            instructions = InstructionSet(resource_requester=dummy_requester)
            instructions.add(instruction, [Qubit()], [])
            bit = Clbit()
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(bit, 0)
            dummy_requester.assert_called_once_with(bit)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(instructions[0].operation.condition[0], sentinel_bit)
        with self.subTest("calls requester with index"):
            dummy_requester.reset_mock()
            instruction = RZGate(0)
            instructions = InstructionSet(resource_requester=dummy_requester)
            instructions.add(instruction, [Qubit()], [])
            index = 0
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(index, 0)
            dummy_requester.assert_called_once_with(index)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(instructions[0].operation.condition[0], sentinel_bit)
        with self.subTest("calls requester with register"):
            dummy_requester.reset_mock()
            instruction = RZGate(0)
            instructions = InstructionSet(resource_requester=dummy_requester)
            instructions.add(instruction, [Qubit()], [])
            register = ClassicalRegister(2)
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(register, 0)
            dummy_requester.assert_called_once_with(register)
            with self.assertWarns(DeprecationWarning):
                self.assertIs(instructions[0].operation.condition[0], sentinel_register)
        with self.subTest("calls requester only once when broadcast"):
            dummy_requester.reset_mock()
            instruction_list = [RZGate(0), RZGate(0), RZGate(0)]
            instructions = InstructionSet(resource_requester=dummy_requester)
            for instruction in instruction_list:
                instructions.add(instruction, [Qubit()], [])
            register = ClassicalRegister(2)
            with self.assertWarns(DeprecationWarning):
                instructions.c_if(register, 0)
            dummy_requester.assert_called_once_with(register)
            for instruction in instruction_list:
                with self.assertWarns(DeprecationWarning):
                    self.assertIs(instructions[0].operation.condition[0], sentinel_register)

    def test_label_type_enforcement(self):
        """Test instruction label type enforcement."""
        with self.subTest("accepts string labels"):
            instruction = Instruction("h", 1, 0, [], label="label")
            self.assertEqual(instruction.label, "label")
        with self.subTest("raises when a non-string label is provided to constructor"):
            with self.assertRaisesRegex(TypeError, r"label expects a string or None"):
                Instruction("h", 1, 0, [], label=0)
        with self.subTest("raises when a non-string label is provided to setter"):
            with self.assertRaisesRegex(TypeError, r"label expects a string or None"):
                instruction = RZGate(0)
                instruction.label = 0


class TestInverseAnnotatedGate(QiskitTestCase):
    """Tests for inverse gates and the AnnotatedOperation class."""

    def test_inverse_cx(self):
        """Test creation of inverse CX gate"""
        gate = CXGate().inverse(annotated=False)
        self.assertIsInstance(gate, CXGate)
        gate = CXGate().inverse(annotated=True)
        self.assertIsInstance(gate, CXGate)

    def test_inverse_s(self):
        """Test creation of inverse S gate"""
        gate = SGate().inverse(annotated=False)
        self.assertIsInstance(gate, SdgGate)
        gate = SGate().inverse(annotated=True)
        self.assertIsInstance(gate, SdgGate)

    def test_inverse_custom(self):
        """Test creation of inverse custom gate"""
        circ = QuantumCircuit(2)
        circ.cx(0, 1)
        circ.h(0)
        gate = circ.to_instruction()
        self.assertIsInstance(gate, Instruction)
        inverse_gate = gate.inverse(annotated=False)
        self.assertIsInstance(inverse_gate, Instruction)
        inverse_gate = gate.inverse(annotated=True)
        self.assertIsInstance(inverse_gate, AnnotatedOperation)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for the methods in ``utils.classtools``."""

# If `utils.wrap_method` is horrendously broken, then the test suite probably won't even have been
# able to collect, because it's used in `QiskitTestCase`.
#
# Throughout this file, we use ``this`` as an argument name to refer to a general reference, which
# may be an instance or the type of the instance.  This is just for clarity when we are using the
# same function to wrap both instance and class methods.
#
# We use a lot of dummy classes in test cases, for which there is absolutely no point defining
# docstrings.
# pylint: disable=missing-class-docstring,missing-function-docstring

import unittest
import sys

from qiskit.utils import wrap_method
from test import QiskitTestCase  # pylint: disable=wrong-import-order


def call_first_argument_with(*args, **kwargs):
    """Create a function that calls its first and only argument with the signature given to this
    function."""

    def out(mock):
        return mock(*args, **kwargs)

    return out


def call_second_argument_with(*args, **kwargs):
    """Create a function that calls its second argument with the first, and the signature given to
    this function."""

    def out(this, mock):
        mock(this, *args, **kwargs)

    return out


class TestWrapMethod(QiskitTestCase):
    """Tests of ``utils.classtools.wrap_method``.  These can be rather tricky, because there's a lot
    of messing around with descriptors and some special cases necessary to support builtin Python
    functions, so we sometimes have rather funny access patterns."""

    def test_called_with(self):
        """Test the basic call patterns are correct.  We use regular Python functions rather than
        mocks to make the instances and callbacks in this simplest case, because the low-level
        descriptor use means that there might be side-effects to binding mocks directly."""

        class Dummy:
            def instance(self, mock):
                mock(self, "method")

            @classmethod
            def class_(cls, mock):
                mock(cls, "method")

            @staticmethod
            def static(mock):
                mock("method")

        wrap_method(
            Dummy,
            "instance",
            before=call_second_argument_with("before"),
            after=call_second_argument_with("after"),
        )
        wrap_method(
            Dummy,
            "class_",
            before=call_second_argument_with("before"),
            after=call_second_argument_with("after"),
        )
        wrap_method(
            Dummy,
            "static",
            before=call_first_argument_with("before"),
            after=call_first_argument_with("after"),
        )

        with self.subTest("from instance"):
            source = Dummy()
            with self.subTest("instance"):
                mock = unittest.mock.Mock()

                # Test that the before/after are not called when the method is accessed.
                caller = source.instance
                mock.assert_not_called()
                caller(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call(source, "before"),
                        unittest.mock.call(source, "method"),
                        unittest.mock.call(source, "after"),
                    ],
                    any_order=False,
                )
            with self.subTest("class"):
                mock = unittest.mock.Mock()
                caller = source.class_
                mock.assert_not_called()
                caller(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call(Dummy, "before"),
                        unittest.mock.call(Dummy, "method"),
                        unittest.mock.call(Dummy, "after"),
                    ],
                    any_order=False,
                )
            with self.subTest("static"):
                mock = unittest.mock.Mock()
                caller = source.static
                mock.assert_not_called()
                caller(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call("before"),
                        unittest.mock.call("method"),
                        unittest.mock.call("after"),
                    ],
                    any_order=False,
                )

        with self.subTest("from type"):
            with self.subTest("instance"):
                mock = unittest.mock.Mock()

                # Test that the before/after are not called when the method is accessed.
                caller = Dummy.instance
                mock.assert_not_called()
                caller("this", mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call("this", "before"),
                        unittest.mock.call("this", "method"),
                        unittest.mock.call("this", "after"),
                    ],
                    any_order=False,
                )
            with self.subTest("class"):
                mock = unittest.mock.Mock()
                caller = Dummy.class_
                mock.assert_not_called()
                caller(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call(Dummy, "before"),
                        unittest.mock.call(Dummy, "method"),
                        unittest.mock.call(Dummy, "after"),
                    ],
                    any_order=False,
                )
            with self.subTest("static"):
                mock = unittest.mock.Mock()
                caller = Dummy.static
                mock.assert_not_called()
                caller(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call("before"),
                        unittest.mock.call("method"),
                        unittest.mock.call("after"),
                    ],
                    any_order=False,
                )

    def test_can_wrap_with_lambda(self):
        """Test that lambda functions can be used as the callbacks."""

        class Dummy:
            def instance(self, mock):
                mock(self, "method")

        wrap_method(Dummy, "instance", after=lambda self, mock: mock(self, "after"))
        with self.subTest("from instance"):
            mock = unittest.mock.Mock()
            source = Dummy()
            caller = source.instance
            mock.assert_not_called()
            caller(mock)
            mock.assert_has_calls(
                [unittest.mock.call(source, "method"), unittest.mock.call(source, "after")],
                any_order=False,
            )

        with self.subTest("from type"):
            mock = unittest.mock.Mock()
            caller = Dummy.instance
            mock.assert_not_called()
            caller("this", mock)
            mock.assert_has_calls(
                [unittest.mock.call("this", "method"), unittest.mock.call("this", "after")],
                any_order=False,
            )

    def test_can_wrap_with_callable_class(self):
        """Test that a class with a ``__call__`` but no descriptor protocol can be used as the
        callbacks."""

        class Dummy:
            def instance(self, mock):
                mock(self, "method")

        class Callback:
            # Note that this class does not implement the descriptor protocol, unlike normal Python
            # functions.  ``__call__`` itself implements ``__get__``, but that just bounds the
            # ``Callback`` instance to ``self``.
            def __call__(self, this, mock):
                mock(this, "after")

        wrap_method(Dummy, "instance", after=Callback())
        with self.subTest("from instance"):
            mock = unittest.mock.Mock()
            source = Dummy()
            caller = source.instance
            mock.assert_not_called()
            caller(mock)
            mock.assert_has_calls(
                [unittest.mock.call(source, "method"), unittest.mock.call(source, "after")],
                any_order=False,
            )

        with self.subTest("from type"):
            mock = unittest.mock.Mock()
            caller = Dummy.instance
            mock.assert_not_called()
            caller("this", mock)
            mock.assert_has_calls(
                [unittest.mock.call("this", "method"), unittest.mock.call("this", "after")],
                any_order=False,
            )

    def test_can_wrap_with_builtin(self):
        """Test that builtin functions can be used a callback.  Many CPython builtins don't
        implement the desriptor protocol that all functions defined with ``def`` or ``lambda`` do,
        which means we need to take special care that they work.  This is most relevant for
        C-extension functions created via pybind11, Cython or similar, rather than actual Python
        builtins."""

        class Dummy:
            def instance(self, mock):
                mock(self, "method")

        # We don't want to add additional compilation requirements for one simple test in the suite,
        # so instead we need a CPython builtin function (of type ``types.BuiltinFunctionType``) that
        # has side effects when being called with an arbitrary object as the first positional
        # argument.  ``breakpoint`` is a good choice, because it's designed to support dependency
        # injection with arbitrary arguments; it calls out to the overridable ``sys.breakpointhook``
        # with all its arguments.  We can't use ``eval`` to test the bound-method calls because the
        # first argument would be the instance, not the string with a side-effect-y programme.
        with unittest.mock.patch.object(sys, "breakpointhook") as mock:
            wrap_method(Dummy, "instance", before=breakpoint)

            with self.subTest("from instance"):
                mock.reset_mock()
                source = Dummy()
                source.instance(mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call(source, mock),
                        unittest.mock.call(source, "method"),
                    ]
                )

            with self.subTest("from type"):
                mock.reset_mock()
                Dummy.instance("this", mock)
                mock.assert_has_calls(
                    [
                        unittest.mock.call("this", mock),
                        unittest.mock.call("this", "method"),
                    ]
                )

    def test_can_wrap_with_mock(self):
        """This is kind of a meta test, to check that we can use a ``unittest.mock.Mock`` instance
        as the callback."""

        class Dummy:
            def instance(self, x):
                pass

        mock = unittest.mock.Mock()
        wrap_method(Dummy, "instance", before=mock)

        with self.subTest("from instance"):
            mock.reset_mock()
            source = Dummy()
            source.instance("hello, world")
            mock.assert_called_once_with(source, "hello, world")
        with self.subTest("from type"):
            mock.reset_mock()
            Dummy.instance("this", "hello, world")
            mock.assert_called_once_with("this", "hello, world")

    def test_wrapping_inherited_method(self):
        """Test that ``wrap_method`` will correctly find a method defined only on a parent class."""

        class Parent:
            def instance(self, mock):
                mock(self, "method")

        class Child(Parent):
            pass

        wrap_method(Child, "instance", before=call_second_argument_with("before"))

        with self.subTest("from instance"):
            mock = unittest.mock.Mock()
            source = Child()
            caller = source.instance
            mock.assert_not_called()
            caller(mock)
            mock.assert_has_calls(
                [unittest.mock.call(source, "before"), unittest.mock.call(source, "method")],
                any_order=False,
            )

        with self.subTest("from type"):
            mock = unittest.mock.Mock()
            caller = Child.instance
            mock.assert_not_called()
            caller("this", mock)
            mock.assert_has_calls(
                [unittest.mock.call("this", "before"), unittest.mock.call("this", "method")],
                any_order=False,
            )

    def test_wrapping___init__(self):
        """Test that wrapping the magic __init__ method works."""

        class Dummy:
            def __init__(self, mock):
                mock("__init__")
                self.mock = mock

        def add_extra_property(self, _):
            mock("extra")
            self.extra = "hello, world"

        wrap_method(Dummy, "__init__", after=add_extra_property)

        mock = unittest.mock.Mock()
        dummy = Dummy(mock)
        self.assertIs(dummy.mock, mock)
        self.assertEqual(dummy.extra, "hello, world")  # pylint: disable=no-member
        mock.assert_has_calls(
            [unittest.mock.call("__init__"), unittest.mock.call("extra")],
            any_order=False,
        )

    def test_wrapping___add__(self):
        """Test that wrapping an arithmetic operator works.  There is nothing particularly special
        about ``__add__`` that (say) ``__init__`` doesn't also do, but this is just a further check
        that the magic methods can work.  Note that ``__add__`` must be defined on the type; all
        magic methods ignore re-definitions in instance dictionaries."""

        mock = unittest.mock.Mock()

        class Dummy:
            def __init__(self, n):
                self.n = n

            def __add__(self, other):
                return type(self)(self.n + other.n)

        wrap_method(Dummy, "__add__", before=mock)

        left = Dummy(1)
        right = Dummy(2)
        out = left + right
        self.assertIsInstance(out, Dummy)
        self.assertEqual(out.n, 3)
        mock.assert_has_calls([unittest.mock.call(left, right)])

    def test_wrapping___new__(self):
        """Test that wrapping the magic __new__ method works.  This method is implicitly made into a
        static method with no decorator, but still gets called with the class in the first
        position.  Note that ``type`` implements ``__new__``, so the getter needs to ensure that it
        doesn't accidentally"""

        class Dummy:
            def __new__(cls, mock):
                mock(cls, "__new__")
                return super().__new__(cls)

        wrap_method(Dummy, "__new__", before=call_second_argument_with("extra"))

        mock = unittest.mock.Mock()
        dummy = Dummy(mock)
        mock.assert_has_calls(
            [unittest.mock.call(Dummy, "extra"), unittest.mock.call(Dummy, "__new__")],
            any_order=False,
        )

    def test_wrapping_object___new__(self):
        """Test that wrapping the magic __new__ method works when it is inherited.  This is a very
        special case, because by inheritance ``A.__new__`` is ``type.__new__``, but that's not we
        want to wrap; we need to have used ``type.__getattribute__`` to make sure that we're getting
        the default implementation ``object.__new__``, and not the ``__new__`` method that literally
        constructs new types.
        """

        class Dummy:
            pass

        mock = unittest.mock.Mock()
        wrap_method(Dummy, "__new__", before=mock)
        dummy = Dummy()
        mock.assert_called_once_with(Dummy)

    def test_wrapping_object___eq__(self):
        """Test that wrapping equality works.  ``type`` also implements ``__eq__`` in a way that
        returns ``NotImplemented`` if one of the operands is not a ``type``, so this tests that we
        are successfully finding ``object.__eq__`` in the resolution of the wrapped method."""

        class Dummy:
            def __init__(self, n):
                self.n = n

            def __eq__(self, other):
                return self.n == other.n

        mock = unittest.mock.Mock()
        wrap_method(Dummy, "__eq__", before=mock)

        left = Dummy(1)
        right = Dummy(2)
        self.assertNotEqual(left, right)
        mock.assert_has_calls([unittest.mock.call(left, right)])

    def test_wrapping_object___init_subclass__(self):
        """Test that wrapping the magic ``__init_subclass__`` method works.  This method is
        implicitly made into a class method without needing a decorator."""

        class Dummy:
            pass

        mock = unittest.mock.Mock()
        wrap_method(Dummy, "__init_subclass__", before=mock)

        class Child(Dummy):
            pass

        mock.assert_called_once_with(Child)

    def test_wrapping_already_wrapped_method(self):
        """Test that a chain of wrapped methods evaluate correctly, in the right order.  Methods
        that are explicitly overridden in child class definitions do not create chains of wrapped
        methods in the same way, even if they call ``super().method`` because the actual object in
        the child definition would be a regular function.  This tests the case that method we're
        wrapping is the exact output of a previous wrapping."""

        class Grandparent:
            def method(self, mock):
                mock(self, "grandparent")

        wrap_method(
            Grandparent,
            "method",
            before=call_second_argument_with("before 1"),
            after=call_second_argument_with("after 1"),
        )

        class Parent(Grandparent):
            pass

        wrap_method(
            Parent,
            "method",
            before=call_second_argument_with("before 2"),
            after=call_second_argument_with("after 2"),
        )

        class Child(Parent):
            pass

        wrap_method(
            Child,
            "method",
            before=call_second_argument_with("before 3"),
            after=call_second_argument_with("after 3"),
        )

        mock = unittest.mock.Mock()
        child = Child()
        child.method(mock)
        mock.assert_has_calls(
            [
                unittest.mock.call(child, "before 3"),
                unittest.mock.call(child, "before 2"),
                unittest.mock.call(child, "before 1"),
                unittest.mock.call(child, "grandparent"),
                unittest.mock.call(child, "after 1"),
                unittest.mock.call(child, "after 2"),
                unittest.mock.call(child, "after 3"),
            ],
            any_order=False,
        )

    def test_docstring_inherited(self):
        """Test that the docstring of a method is correctly passed through, to avoid clobbering
        documentation."""

        class Dummy:
            def method(self):
                """This is documentation."""

        wrap_method(Dummy, "method", before=lambda self: None)
        self.assertEqual(Dummy.method.__doc__, "This is documentation.")

    def test_raises_on_invalid_name(self):
        """Test that a suitable error is raised if the method doesn't exist."""

        class Dummy:
            pass

        with self.assertRaisesRegex(AttributeError, "bad"):
            wrap_method(Dummy, "bad", before=lambda self: None)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test library of graph state circuits."""

import unittest
import numpy as np

from qiskit.circuit import Gate
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit.library import GraphState, GraphStateGate
from qiskit.quantum_info import Clifford, Operator
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestGraphStateLibrary(QiskitTestCase):
    """Test the graph state circuit."""

    def assertGraphStateIsCorrect(self, adjacency_matrix, graph_state):
        """Check the stabilizers of the graph state circuit/gate against the expected stabilizers.
        Based on https://arxiv.org/pdf/quant-ph/0307130.pdf, Eq. (6).
        """
        if isinstance(graph_state, Gate):
            cliff = Clifford(graph_state.definition)
        else:
            cliff = Clifford(graph_state)

        stabilizers = [stabilizer[1:] for stabilizer in cliff.to_labels(mode="S")]

        expected_stabilizers = []  # keep track of all expected stabilizers
        num_vertices = len(adjacency_matrix)
        for vertex_a in range(num_vertices):
            stabilizer = [None] * num_vertices  # Paulis must be put into right place
            for vertex_b in range(num_vertices):
                if vertex_a == vertex_b:  # self-connection --> 'X'
                    stabilizer[vertex_a] = "X"
                elif adjacency_matrix[vertex_a][vertex_b] != 0:  # vertices connected --> 'Z'
                    stabilizer[vertex_b] = "Z"
                else:  # else --> 'I'
                    stabilizer[vertex_b] = "I"

            # need to reverse for Qiskit's tensoring order
            expected_stabilizers.append("".join(stabilizer)[::-1])

        self.assertListEqual(expected_stabilizers, stabilizers)

    def test_graph_state_circuit(self):
        """Verify the GraphState by checking if the circuit has the expected stabilizers."""
        adjacency_matrix = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        graph_state = GraphState(adjacency_matrix)
        self.assertGraphStateIsCorrect(adjacency_matrix, graph_state)

    def test_non_symmetric_circuit_raises(self):
        """Test that adjacency matrix is required to be symmetric."""
        adjacency_matrix = [[1, 1, 0], [1, 0, 1], [1, 1, 1]]
        with self.assertRaises(CircuitError):
            GraphState(adjacency_matrix)

    def test_graph_state_gate(self):
        """Verify correctness of GraphStatGate by checking that the gate's definition circuit
        has the expected stabilizers.
        """
        adjacency_matrix = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        graph_state = GraphStateGate(adjacency_matrix)
        self.assertGraphStateIsCorrect(adjacency_matrix, graph_state)

    def test_non_symmetric_gate_raises(self):
        """Test that adjacency matrix is required to be symmetric."""
        adjacency_matrix = [[1, 1, 0], [1, 0, 1], [1, 1, 1]]
        with self.assertRaises(CircuitError):
            GraphStateGate(adjacency_matrix)

    def test_circuit_and_gate_equivalence(self):
        """Test that GraphState-circuit and GraphStateGate yield equal operators."""
        adjacency_matrix = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        graph_state_gate = GraphStateGate(adjacency_matrix)
        graph_state_circuit = GraphState(adjacency_matrix)
        self.assertEqual(Operator(graph_state_gate), Operator(graph_state_circuit))

    def test_adjacency_matrix(self):
        """Test GraphStateGate's adjacency_matrix method."""
        adjacency_matrix = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        graph_state_gate = GraphStateGate(adjacency_matrix)
        self.assertTrue(np.all(graph_state_gate.adjacency_matrix == adjacency_matrix))

    def test_equality(self):
        """Test GraphStateGate's equality method."""
        mat1 = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        mat2 = [
            [0, 1, 0, 0, 1],
            [1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        with self.subTest("equal"):
            gate1 = GraphStateGate(mat1)
            gate2 = GraphStateGate(mat1)
            self.assertEqual(gate1, gate2)
        with self.subTest("not equal"):
            gate1 = GraphStateGate(mat1)
            gate2 = GraphStateGate(mat2)
            self.assertNotEqual(gate1, gate2)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Anti commutator function."""

from __future__ import annotations
from typing import TypeVar

from qiskit.quantum_info.operators.linear_op import LinearOp

OperatorTypeT = TypeVar("OperatorTypeT", bound=LinearOp)


def anti_commutator(a: OperatorTypeT, b: OperatorTypeT) -> OperatorTypeT:
    r"""Compute anti-commutator of a and b.

    .. math::

        ab + ba.

    Args:
        a: Operator a.
        b: Operator b.
    Returns:
        The anti-commutator
    """
    return a @ b + b @ a



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Code from commutative_analysis pass that checks commutation relations between DAG nodes."""

from typing import List, Union, Set, Optional

from qiskit.circuit.operation import Operation
from qiskit._accelerate.commutation_checker import CommutationChecker as RustChecker


class CommutationChecker:
    """This code is essentially copy-pasted from commutative_analysis.py.
    This code cleverly hashes commutativity and non-commutativity results between DAG nodes and seems
    quite efficient for large Clifford circuits.
    They may be other possible efficiency improvements: using rule-based commutativity analysis,
    evicting from the cache less useful entries, etc.
    """

    def __init__(
        self,
        standard_gate_commutations: dict = None,
        cache_max_entries: int = 10**6,
        *,
        gates: Optional[Set[str]] = None,
    ):
        self.cc = RustChecker(standard_gate_commutations, cache_max_entries, gates)

    def commute_nodes(
        self,
        op1,
        op2,
        max_num_qubits: int = 3,
    ) -> bool:
        """Checks if two DAGOpNodes commute."""
        return self.cc.commute_nodes(op1, op2, max_num_qubits)

    def commute(
        self,
        op1: Operation,
        qargs1: List,
        cargs1: List,
        op2: Operation,
        qargs2: List,
        cargs2: List,
        max_num_qubits: int = 3,
    ) -> bool:
        """
        Checks if two Operations commute. The return value of `True` means that the operations
        truly commute, and the return value of `False` means that either the operations do not
        commute or that the commutation check was skipped (for example, when the operations
        have conditions or have too many qubits).

        Args:
            op1: first operation.
            qargs1: first operation's qubits.
            cargs1: first operation's clbits.
            op2: second operation.
            qargs2: second operation's qubits.
            cargs2: second operation's clbits.
            max_num_qubits: the maximum number of qubits to consider, the check may be skipped if
                the number of qubits for either operation exceeds this amount.

        Returns:
            bool: whether two operations commute.
        """
        return self.cc.commute(op1, qargs1, cargs1, op2, qargs2, cargs2, max_num_qubits)

    def num_cached_entries(self):
        """Returns number of cached entries"""
        return self.cc.num_cached_entries()

    def clear_cached_commutations(self):
        """Clears the dictionary holding cached commutations"""
        self.cc.clear_cached_commutations()

    def check_commutation_entries(
        self,
        first_op: Operation,
        first_qargs: List,
        second_op: Operation,
        second_qargs: List,
    ) -> Union[bool, None]:
        """Returns stored commutation relation if any

        Args:
            first_op: first operation.
            first_qargs: first operation's qubits.
            second_op: second operation.
            second_qargs: second operation's qubits.

        Return:
            bool: True if the gates commute and false if it is not the case.
        """
        return self.cc.library.check_commutation_entries(
            first_op, first_qargs, second_op, second_qargs
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Primitive job abstract base class
"""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

from ..containers import PrimitiveResult
from .base_result import _BasePrimitiveResult

ResultT = TypeVar("ResultT", bound=Union[_BasePrimitiveResult, PrimitiveResult])
StatusT = TypeVar("StatusT")


class BasePrimitiveJob(ABC, Generic[ResultT, StatusT]):
    """Primitive job abstract base class."""

    def __init__(self, job_id: str, **kwargs) -> None:
        """Initializes the primitive job.

        Args:
            job_id: A unique id in the context of the primitive used to run the job.
            kwargs: Any key value metadata to associate with this job.
        """
        self._job_id = job_id
        self.metadata = kwargs

    def job_id(self) -> str:
        """Return a unique id identifying the job."""
        return self._job_id

    @abstractmethod
    def result(self) -> ResultT:
        """Return the results of the job."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `result` method.")

    @abstractmethod
    def status(self) -> StatusT:
        """Return the status of the job."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `status` method.")

    @abstractmethod
    def done(self) -> bool:
        """Return whether the job has successfully run."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `done` method.")

    @abstractmethod
    def running(self) -> bool:
        """Return whether the job is actively running."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `running` method.")

    @abstractmethod
    def cancelled(self) -> bool:
        """Return whether the job has been cancelled."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `cancelled` method.")

    @abstractmethod
    def in_final_state(self) -> bool:
        """Return whether the job is in a final job state such as ``DONE`` or ``ERROR``."""
        raise NotImplementedError(
            "Subclass of BasePrimitiveJob must implement `is_final_state` method."
        )

    @abstractmethod
    def cancel(self):
        """Attempt to cancel the job."""
        raise NotImplementedError("Subclass of BasePrimitiveJob must implement `cancel` method.")
